# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Iterable

import numpy as np

from braket.default_simulator.observables import Hermitian, Identity, TensorProduct
from braket.default_simulator.operation import GateOperation, KrausOperation
from braket.default_simulator.result_types import _from_braket_observable
from braket.ir.jaqcd.program_v1 import Results
from braket.ir.jaqcd.shared_models import Observable, OptionalMultiTarget


class Circuit:
    """
    This is a lightweight analog to braket.ir.jaqcd.program_v1.Program.
    The Interpreter compiles to an IR to hand off to the simulator,
    braket.default_simulator.state_vector_simulator.StateVectorSimulator, for example.
    Our simulator module takes in a circuit specification that satisfies the interface
    implemented by this class.
    """

    def __init__(
        self,
        instructions: list[GateOperation] | None = None,
        results: list[Results] | None = None,
    ):
        self.instructions = []
        self.results = []
        self.qubit_set = set()
        self.measured_qubits = []
        self.target_classical_indices = []

        if instructions:
            for instruction in instructions:
                self.add_instruction(instruction)

        if results:
            for result in results:
                self.add_result(result)

    def add_instruction(self, instruction: [GateOperation, KrausOperation]) -> None:
        """
        Add instruction to the circuit.

        Args:
            instruction (GateOperation): Instruction to add.
        """
        self.instructions.append(instruction)
        self.qubit_set |= set(instruction.targets)

    def add_measure(self, target: tuple[int], classical_targets: Iterable[int] = None):
        for index, qubit in enumerate(target):
            if qubit in self.measured_qubits:
                raise ValueError(f"Qubit {qubit} is already measured or captured.")
            self.measured_qubits.append(qubit)
            self.qubit_set.add(qubit)
            self.target_classical_indices.append(
                classical_targets[index]
                if classical_targets
                else max(index, len(self.target_classical_indices))
            )

    def add_result(self, result: Results) -> None:
        """
        Add result type to the circuit.

        Args:
            result (Results): Result type to add.
        """
        self.results.append(result)
        if isinstance(result, OptionalMultiTarget) and result.targets is not None:
            self.qubit_set |= set(result.targets)

    @property
    def num_qubits(self) -> int:
        return len(self.qubit_set)

    @property
    def basis_rotation_instructions(self):
        """Basis rotation instructions implied by the provided observables"""
        basis_rotation_instructions = []
        observable_map = {}

        def process_observable(observable):
            if isinstance(observable, Identity):
                return
            measured_qubits = tuple(observable.measured_qubits)
            for qubit in measured_qubits:
                for target, previously_measured in observable_map.items():
                    if qubit in target:
                        # must ensure that target is the same
                        if target != measured_qubits:
                            raise ValueError("Qubit part of incompatible results targets")
                        # must ensure observable is the same
                        if type(previously_measured) is not type(observable):
                            raise ValueError("Conflicting result types applied to a single qubit")
                        # including matrix value for Hermitians
                        if isinstance(observable, Hermitian):
                            if not np.allclose(previously_measured.matrix, observable.matrix):
                                raise ValueError(
                                    "Conflicting result types applied to a single qubit"
                                )
            observable_map[measured_qubits] = observable

        for result in self.results:
            if isinstance(result, Observable):
                observables = result.observable

                if result.targets is not None:
                    braket_obs = _from_braket_observable(observables, result.targets)

                    if isinstance(braket_obs, TensorProduct):
                        for factor in braket_obs.factors:
                            process_observable(factor)
                    else:
                        process_observable(braket_obs)

                else:
                    for q in range(self.num_qubits):
                        braket_obs = _from_braket_observable(observables, [q])
                        process_observable(braket_obs)

        for obs in observable_map.values():
            diagonalizing_gates = obs.diagonalizing_gates(self.num_qubits)
            basis_rotation_instructions.extend(diagonalizing_gates)

        return basis_rotation_instructions

    def __eq__(self, other: Circuit):
        return (self.instructions, self.results) == (other.instructions, other.results)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import math
from enum import Enum

import numpy as np
import sympy

from braket.default_simulator.openqasm._helpers.casting import cast_to
from braket.default_simulator.openqasm.parser.openqasm_ast import (
    ArrayLiteral,
    BitstringLiteral,
    FloatLiteral,
    IntegerLiteral,
    SymbolLiteral,
    UintType,
)


class BuiltinConstants(Enum):
    PI = SymbolLiteral(sympy.pi)
    TAU = SymbolLiteral(2 * sympy.pi)
    E = SymbolLiteral(sympy.E)


class BuiltinFunctions:
    @staticmethod
    def sizeof(array: ArrayLiteral, dim: IntegerLiteral | None = None) -> IntegerLiteral:
        return (
            IntegerLiteral(len(array.values))
            if dim is None or dim.value == 0
            else BuiltinFunctions.sizeof(array.values[0], IntegerLiteral(dim.value - 1))
        )

    @staticmethod
    def arccos(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.acos(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.arccos(x.value))
        )

    @staticmethod
    def arcsin(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.asin(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.arcsin(x.value))
        )

    @staticmethod
    def arctan(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.atan(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.arctan(x.value))
        )

    @staticmethod
    def ceiling(x: FloatLiteral | SymbolLiteral) -> IntegerLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.ceiling(x.value))
            if isinstance(x, SymbolLiteral)
            else IntegerLiteral(math.ceil(x.value))
        )

    @staticmethod
    def cos(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.cos(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.cos(x.value))
        )

    @staticmethod
    def exp(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.exp(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.exp(x.value))
        )

    @staticmethod
    def floor(x: FloatLiteral | SymbolLiteral) -> IntegerLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.floor(x.value))
            if isinstance(x, SymbolLiteral)
            else IntegerLiteral(np.floor(x.value))
        )

    @staticmethod
    def log(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.log(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.log(x.value))
        )

    @staticmethod
    def mod(
        x: IntegerLiteral | FloatLiteral | SymbolLiteral,
        y: IntegerLiteral | FloatLiteral | SymbolLiteral,
    ) -> IntegerLiteral | FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(x.value % y.value)
            if isinstance(x, SymbolLiteral) or isinstance(y, SymbolLiteral)
            else (
                IntegerLiteral(x.value % y.value)
                if isinstance(x, IntegerLiteral) and isinstance(y, IntegerLiteral)
                else FloatLiteral(x.value % y.value)
            )
        )

    @staticmethod
    def popcount(x: IntegerLiteral | BitstringLiteral | ArrayLiteral) -> IntegerLiteral:
        # does not support symbols or expressions
        return IntegerLiteral(np.binary_repr(cast_to(UintType(), x).value).count("1"))

    @staticmethod
    def pow(
        x: IntegerLiteral | FloatLiteral | SymbolLiteral,
        y: IntegerLiteral | FloatLiteral | SymbolLiteral,
    ) -> IntegerLiteral | FloatLiteral | SymbolLiteral:
        # parser gets confused by pow, mistaking for quantum modifier
        return (
            SymbolLiteral(x.value**y.value)
            if isinstance(x, SymbolLiteral) or isinstance(y, SymbolLiteral)
            else (
                IntegerLiteral(x.value**y.value)
                if isinstance(x, IntegerLiteral) and isinstance(y, IntegerLiteral)
                else FloatLiteral(x.value**y.value)
            )
        )

    @staticmethod
    def rotl(x, y):
        raise NotImplementedError

    @staticmethod
    def rotr(x, y):
        raise NotImplementedError

    @staticmethod
    def sin(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.sin(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.sin(x.value))
        )

    @staticmethod
    def sqrt(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.sqrt(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.sqrt(x.value))
        )

    @staticmethod
    def tan(x: FloatLiteral | SymbolLiteral) -> FloatLiteral | SymbolLiteral:
        return (
            SymbolLiteral(sympy.tan(x.value))
            if isinstance(x, SymbolLiteral)
            else FloatLiteral(np.tan(x.value))
        )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import singledispatch
from typing import Any

import numpy as np

from braket.default_simulator.linalg_utils import marginal_probability, partial_trace
from braket.default_simulator.observables import (
    Hadamard,
    Hermitian,
    Identity,
    PauliX,
    PauliY,
    PauliZ,
    TensorProduct,
)
from braket.default_simulator.operation import Observable
from braket.default_simulator.operation_helpers import ir_matrix_to_ndarray
from braket.default_simulator.simulation import Simulation
from braket.default_simulator.state_vector_simulation import StateVectorSimulation
from braket.ir import jaqcd


def from_braket_result_type(result_type) -> ResultType:
    """Creates a `ResultType` corresponding to the given Braket instruction.

    Args:
        result_type: Result type for a circuit specified using the `braket.ir.jacqd` format.

    Returns:
        ResultType: Instance of specific `ResultType` corresponding to the type of result_type

    Raises:
        NotImplementedError: If no concrete `ResultType` class has been registered
            for the Braket instruction type
    """
    return _from_braket_result_type(result_type)


@singledispatch
def _from_braket_result_type(result_type):
    raise NotImplementedError(f"Result type {result_type} not recognized")


class ResultType(ABC):
    """
    An abstract class that when implemented defines a calculation on a
    quantum state simulation.

    Note:
        All result types are calculated exactly, instead of approximated from samples.
        Sampled results are returned from `Simulation.retrieve_samples`, which can be processed by,
        for example, the Amazon Braket SDK.
    """

    @abstractmethod
    def calculate(self, simulation: Simulation) -> Any:
        # Return type of Any due to lack of sum type support in Python
        """Calculate a result from the given quantum state vector simulation.

        Args:
            simulation (Simulation): The simulation to use in the calculation.

        Returns:
            Any: The result of the calculation.
        """


class TargetedResultType(ResultType, ABC):
    """
    Holds an observable that may target qubits.
    """

    def __init__(self, targets: list[int] | None = None):
        """
        Args:
            targets (list[int], optional): The target qubits of the result type.
                If None, no specific qubits are targeted.
        """
        self._targets = targets

    @property
    def targets(self) -> tuple[int, ...] | None:
        """tuple[int], optional: The target qubits of the result type, if any."""
        return self._targets


class ObservableResultType(TargetedResultType, ABC):
    """
    Holds an observable to perform a calculation in conjunction with a state.
    """

    def __init__(self, observable: Observable):
        """
        Args:
            observable (Observable): The observable for which the desired result is calculated
        """
        super().__init__(observable.measured_qubits)
        self._observable = observable

    @property
    def observable(self):
        """Observable: The observable for which the desired result is calculated."""
        return self._observable

    @property
    def targets(self) -> tuple[int, ...] | None:
        return self._observable.measured_qubits

    def calculate(self, simulation: Simulation) -> float | list[float]:
        """Calculates the result type using the underlying observable.

        Returns a real number if the observable has defined targets,
        or a list of real numbers, one for the result type on each target,
        if the observable has no target.

        Args:
            simulation (Simulation): The simulation to use in the calculation.

        Returns:
            float | list[float]: The value of the result type;
            will be a real due to self-adjointness of observable.
        """
        if self._observable.measured_qubits:
            return self._calculate_single_quantity(simulation, self._observable)
        return [
            self._calculate_single_quantity(simulation, self._observable.fix_qubit(qubit))
            for qubit in range(simulation.qubit_count)
        ]

    @staticmethod
    @abstractmethod
    def _calculate_single_quantity(simulation: Simulation, observable: Observable) -> float:
        """Calculates a single real value of the result type.

        Args:
            simulation (Simulation): The simulation to use in the calculation.
            observable (Observable): The observable used to calculate the result type.

        Returns:
            float: The value of the result type.
        """


class StateVector(ResultType):
    """
    Simply returns the given state vector.
    """

    def calculate(self, simulation: StateVectorSimulation) -> np.ndarray:
        """Return the given state vector of the simulation.

        Args:
            simulation (StateVectorSimulation): The simulation whose state vector will be returned

        Returns:
            np.ndarray: The state vector (before observables) of the simulation
        """
        return simulation.state_vector


@_from_braket_result_type.register
def _(_: jaqcd.StateVector):
    return StateVector()


class DensityMatrix(TargetedResultType):
    """
    Simply returns the given density matrix.
    """

    def __init__(self, targets: list[int] | None = None):
        """
        Args:
            targets (list[int] | None): The qubit indices on which the reduced density matrix
                are desired. If no targets are specified, the full density matrix is calculated.
                Default: `None`
        """
        super().__init__(targets)

    def calculate(self, simulation: Simulation) -> np.ndarray:
        """Return the given density matrix of the simulation.

        Args:
            simulation (Simulation): The simulation whose (full or reduced)
                density matrix will be returned.

        Returns:
            np.ndarray: The density matrix (before observables) of the simulation
        """
        if self._targets is None or np.array_equal(self._targets, range(simulation.qubit_count)):
            return simulation.density_matrix
        else:
            if not all(ta in list(range(simulation.qubit_count)) for ta in self._targets):
                raise IndexError(
                    "Input target qubits must be within the range of the qubits in the circuit."
                )
            return partial_trace(
                simulation.density_matrix.reshape(np.array([2] * 2 * simulation.qubit_count)),
                self._targets,
            )


@_from_braket_result_type.register
def _(density_matrix: jaqcd.DensityMatrix):
    return DensityMatrix(density_matrix.targets)


class Amplitude(ResultType):
    """
    Extracts the amplitudes of the desired computational basis states.
    """

    def __init__(self, states: list[str]):
        """
        Args:
            states (list[str]): The computational basis states whose amplitudes are desired
        """
        self._states = states

    def calculate(self, simulation: StateVectorSimulation) -> dict[str, complex]:
        """Return the amplitudes of the desired computational basis states in the state
        of the given simulation.

        Args:
            simulation (StateVectorSimulation): The simulation whose state vector amplitudes
                will be returned

        Returns:
            dict[str, complex]: A dict keyed on computational basis states as bitstrings,
            with corresponding values the amplitudes
        """
        state = simulation.state_vector
        return {basis_state: state[int(basis_state, 2)] for basis_state in self._states}


@_from_braket_result_type.register
def _(amplitude: jaqcd.Amplitude):
    return Amplitude(amplitude.states)


class Probability(TargetedResultType):
    """
    Computes the marginal probabilities of computational basis states on the desired qubits.
    """

    def __init__(self, targets: list[int] | None = None):
        """
        Args:
            targets (list[int] | None): The qubit indices on which probabilities are desired.
                If no targets are specified, the probabilities are calculated on the entire state.
                Default: `None`
        """
        super().__init__(targets)

    def calculate(self, simulation: Simulation) -> np.ndarray:
        """Return the marginal probabilities of computational basis states on the target qubits.

        Probabilities are marginalized over all non-target qubits.

        Args:
            simulation (Simulation): The simulation from which probabilities are calculated.

        Returns:
            np.ndarray: An array of probabilities of length equal to 2^(number of target qubits),
            indexed by the decimal encoding of the computational basis state on the target qubits

        """
        return marginal_probability(
            simulation.probabilities,
            self._targets,
        )


@_from_braket_result_type.register
def _(probability: jaqcd.Probability):
    return Probability(probability.targets)


class Expectation(ObservableResultType):
    """
    Holds an observable :math:`O` to calculate its expected value.
    """

    @staticmethod
    def _calculate_single_quantity(simulation: Simulation, observable: Observable) -> float:
        return simulation.expectation(observable)


@_from_braket_result_type.register
def _(expectation: jaqcd.Expectation):
    return Expectation(_from_braket_observable(expectation.observable, expectation.targets))


class Variance(ObservableResultType):
    """
    Holds an observable :math:`O` to calculate its variance.
    """

    @staticmethod
    def _calculate_single_quantity(simulation: Simulation, observable: Observable) -> float:
        return simulation.expectation(observable**2) - simulation.expectation(observable) ** 2


@_from_braket_result_type.register
def _(variance: jaqcd.Variance):
    return Variance(_from_braket_observable(variance.observable, variance.targets))


def _from_braket_observable(
    ir_observable: list[str | list[list[list[float]]]], ir_targets: list[int] | None = None
) -> Observable:
    targets = list(ir_targets) if ir_targets else None
    if len(ir_observable) == 1:
        return _from_single_observable(ir_observable[0], targets)
    else:
        observable = TensorProduct(
            [_from_single_observable(factor, targets, is_factor=True) for factor in ir_observable]
        )
        if targets:
            raise ValueError(
                f"Found {len(targets)} more target qubits than the tensor product acts on"
            )
        return observable


def _from_single_observable(
    observable: str | list[list[list[float]]],
    targets: list[int] | None = None,
    # IR tensor product observables are decoupled from targets
    is_factor: bool = False,
) -> Observable:
    match observable:
        case "i":
            return Identity(_actual_targets(targets, 1, is_factor))
        case "h":
            return Hadamard(_actual_targets(targets, 1, is_factor))
        case "x":
            return PauliX(_actual_targets(targets, 1, is_factor))
        case "y":
            return PauliY(_actual_targets(targets, 1, is_factor))
        case "z":
            return PauliZ(_actual_targets(targets, 1, is_factor))
        case _:
            try:
                matrix = ir_matrix_to_ndarray(observable)
                if is_factor:
                    num_qubits = int(np.log2(len(matrix)))
                    return Hermitian(matrix, _actual_targets(targets, num_qubits, True))
                return Hermitian(matrix, targets)
            except Exception:
                raise ValueError(f"Invalid observable specified: {observable}, targets: {targets}")


def _actual_targets(targets: list[int], num_qubits: int, is_factor: bool):
    if not is_factor:
        return targets
    try:
        return [targets.pop(0) for _ in range(num_qubits)]
    except Exception:
        raise ValueError("Insufficient target qubits for tensor product")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np

from braket.default_simulator.operation import GateOperation, Observable
from braket.default_simulator.simulation import Simulation
from braket.default_simulator.simulation_strategies import (
    batch_operation_strategy,
    single_operation_strategy,
)


class StateVectorSimulation(Simulation):
    """
    This class tracks the evolution of a quantum system with `qubit_count` qubits.
    The state of system the evolves by application of `GateOperation`s using the `evolve()` method.

    How operations are applied is determined by the `batch_size` argument in `__init__`.

    If `batch_size` is set to 1, operations are applied one at a time.

    If `batch_size` is greater than 1, the operation list is first partitioned into a sequence of
    contiguous batches, each of size `batch_size`. If the number of operations is not evenly
    divisible by `batch_size`, then the number of operations in the last batch will just be the
    remainder. The operations in each batch are then applied (contracted) together. The order of the
    operations in the batches are the same as the original order of the operations. In most cases,
    tasks complete faster when run on a larger batch, but require more memory. For more details, see
    the module `batch_operation_strategy`.
    """

    def __init__(self, qubit_count: int, shots: int, batch_size: int):
        r"""
        Args:
            qubit_count (int): The number of qubits being simulated.
                All the qubits start in the :math:`\ket{\mathbf{0}}` computational basis state.
            shots (int): The number of samples to take from the simulation.
                If set to 0, only results that do not require sampling, such as state vector
                or expectation, are generated.
            batch_size (int): The size of the partitions to contract; if set to 1,
                the gates are applied one at a time, without any optimization of
                contraction order. Must be a positive integer.
        """
        if not isinstance(batch_size, int):
            raise TypeError(f"batch_size must be of type `int`, but {type(batch_size)} provided")
        if batch_size < 1:
            raise ValueError(f"batch_size must be a positive integer, but {batch_size} provided")

        super().__init__(qubit_count=qubit_count, shots=shots)
        initial_state = np.zeros(2**qubit_count, dtype=complex)
        initial_state[0] = 1
        self._state_vector = initial_state
        self._batch_size = batch_size
        self._post_observables = None
        self._rng_generator = np.random.default_rng()

    def evolve(self, operations: list[GateOperation]) -> None:
        self._state_vector = StateVectorSimulation._apply_operations(
            self._state_vector, self._qubit_count, operations, self._batch_size
        )

    def apply_observables(self, observables: list[Observable]) -> None:
        """Applies the diagonalizing matrices of the given observables
        to the state of the simulation.

        This method can only be called once.

        Args:
            observables (list[Observable]): The observables to apply

        Raises:
            RuntimeError: If this method is called more than once
        """
        if self._post_observables is not None:
            raise RuntimeError("Observables have already been applied.")
        operations = list(
            sum(
                [observable.diagonalizing_gates(self._qubit_count) for observable in observables],
                (),
            )
        )
        self._post_observables = StateVectorSimulation._apply_operations(
            self._state_vector, self._qubit_count, operations, self._batch_size
        )

    @staticmethod
    def _apply_operations(
        state: np.ndarray, qubit_count: int, operations: list[GateOperation], batch_size: int
    ) -> np.ndarray:
        state_tensor = np.reshape(state, [2] * qubit_count)
        final = (
            single_operation_strategy.apply_operations(state_tensor, qubit_count, operations)
            if batch_size == 1
            else batch_operation_strategy.apply_operations(
                state_tensor, qubit_count, operations, batch_size
            )
        )
        return np.reshape(final, 2**qubit_count)

    def retrieve_samples(self) -> np.ndarray:
        return np.searchsorted(
            np.cumsum(self.probabilities), self._rng_generator.random(size=self._shots)
        )

    @property
    def state_vector(self) -> np.ndarray:
        """
        np.ndarray: The state vector specifying the current state of the simulation.

        Note:
            Mutating this array will mutate the state of the simulation.
        """
        return self._state_vector

    @property
    def density_matrix(self) -> np.ndarray:
        """
        np.ndarray: The density matrix specifying the current state of the simulation.
        """
        return np.outer(self._state_vector, self._state_vector.conj())

    @property
    def state_with_observables(self) -> np.ndarray:
        """
        np.ndarray: The state vector diagonalized in the basis of the measured observables.

        Raises:
            RuntimeError: If observables have not been applied
        """
        if self._post_observables is None:
            raise RuntimeError("No observables applied")
        return self._post_observables

    def expectation(self, observable: Observable) -> float:
        qubit_count = self._qubit_count
        with_observables = observable.apply(np.reshape(self._state_vector, [2] * qubit_count))
        return complex(
            np.dot(self._state_vector.conj(), np.reshape(with_observables, 2**qubit_count))
        ).real

    @property
    def probabilities(self) -> np.ndarray:
        """
        np.ndarray: The probabilities of each computational basis state of the current state
            vector of the simulation.
        """
        return self._probabilities(self.state_vector)

    @staticmethod
    def _probabilities(state: np.ndarray) -> np.ndarray:
        """The probabilities of each computational basis state of a given state vector.

        Args:
            state (np.ndarray): The state vector from which probabilities are extracted.

        Returns:
            np.ndarray: The probabilities of each computational basis state.
        """
        return np.abs(state) ** 2



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Sequence

import numpy as np

import braket.ir.jaqcd as braket_instruction
from braket.default_simulator.operation import GateOperation
from braket.default_simulator.operation_helpers import (
    _from_braket_instruction,
    check_matrix_dimensions,
    check_unitary,
    ir_matrix_to_ndarray,
)


class Identity(GateOperation):
    """Identity gate"""

    _matrix = np.eye(2, dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Identity._matrix

    @property
    def gate_type(self) -> str:
        return "identity"


@_from_braket_instruction.register(braket_instruction.I)
def _i(instruction) -> Identity:
    return Identity([instruction.target])


class Hadamard(GateOperation):
    """Hadamard gate"""

    _matrix = np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Hadamard._matrix

    @property
    def gate_type(self) -> str:
        return "hadamard"


@_from_braket_instruction.register(braket_instruction.H)
def _hadamard(instruction) -> Hadamard:
    return Hadamard([instruction.target])


class PauliX(GateOperation):
    """Pauli-X gate"""

    _matrix = np.array([[0, 1], [1, 0]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return PauliX._matrix

    @property
    def gate_type(self) -> str:
        return "pauli_x"


@_from_braket_instruction.register(braket_instruction.X)
def _pauli_x(instruction) -> PauliX:
    return PauliX([instruction.target])


class PauliY(GateOperation):
    """Pauli-Y gate"""

    _matrix = np.array([[0, -1j], [1j, 0]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return PauliY._matrix

    @property
    def gate_type(self) -> str:
        return "pauli_y"


@_from_braket_instruction.register(braket_instruction.Y)
def _pauli_y(instruction) -> PauliY:
    return PauliY([instruction.target])


class PauliZ(GateOperation):
    """Pauli-Z gate"""

    _matrix = np.diag([1.0, -1.0]).astype(complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return PauliZ._matrix

    @property
    def gate_type(self) -> str:
        return "pauli_z"


@_from_braket_instruction.register(braket_instruction.Z)
def _pauli_z(instruction) -> PauliZ:
    return PauliZ([instruction.target])


class CV(GateOperation):
    """Controlled-Sqrt(NOT) gate"""

    _matrix = np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0.5 + 0.5j, 0.5 - 0.5j],
            [0, 0, 0.5 - 0.5j, 0.5 + 0.5j],
        ],
        dtype=complex,
    )

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return CV._matrix

    @property
    def gate_type(self) -> str:
        return "cv"


@_from_braket_instruction.register(braket_instruction.CV)
def _cv(instruction) -> CV:
    return CV([instruction.control, instruction.target])


class CX(GateOperation):
    """Controlled Pauli-X gate"""

    _matrix = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return CX._matrix

    @property
    def gate_type(self) -> str:
        return "cx"


@_from_braket_instruction.register(braket_instruction.CNot)
def _cx(instruction) -> CX:
    return CX([instruction.control, instruction.target])


class CY(GateOperation):
    """Controlled Pauli-Y gate"""

    _matrix = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self):
        return CY._matrix

    @property
    def gate_type(self) -> str:
        return "cy"


@_from_braket_instruction.register(braket_instruction.CY)
def _cy(instruction) -> CY:
    return CY([instruction.control, instruction.target])


class CZ(GateOperation):
    """Controlled Pauli-Z gate"""

    _matrix = np.diag([1, 1, 1, -1]).astype(complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return CZ._matrix

    @property
    def gate_type(self) -> str:
        return "cz"


@_from_braket_instruction.register(braket_instruction.CZ)
def _cz(instruction) -> CZ:
    return CZ([instruction.control, instruction.target])


class ECR(GateOperation):
    """ECR gate"""

    _matrix = (
        1
        / np.sqrt(2)
        * np.array(
            [[0, 0, 1, 1.0j], [0, 0, 1.0j, 1], [1, -1.0j, 0, 0], [-1.0j, 1, 0, 0]],
            dtype=complex,
        )
    )

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return ECR._matrix

    @property
    def gate_type(self) -> str:
        return "ecr"


@_from_braket_instruction.register(braket_instruction.ECR)
def _ecr(instruction) -> ECR:
    return ECR(instruction.targets)


class S(GateOperation):
    """S gate"""

    _matrix = np.array([[1, 0], [0, 1j]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return S._matrix

    @property
    def gate_type(self) -> str:
        return "s"


@_from_braket_instruction.register(braket_instruction.S)
def _s(instruction) -> S:
    return S([instruction.target])


class Si(GateOperation):
    r"""The adjoint :math:`S^{\dagger}` of the S gate"""

    _matrix = np.array([[1, 0], [0, -1j]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Si._matrix

    @property
    def gate_type(self) -> str:
        return "si"


@_from_braket_instruction.register(braket_instruction.Si)
def _si(instruction) -> Si:
    return Si([instruction.target])


class T(GateOperation):
    """T gate"""

    _matrix = np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return T._matrix

    @property
    def gate_type(self) -> str:
        return "t"


@_from_braket_instruction.register(braket_instruction.T)
def _t(instruction) -> T:
    return T([instruction.target])


class Ti(GateOperation):
    r"""The adjoint :math:`T^{\dagger}` of the T gate"""

    _matrix = np.array([[1, 0], [0, np.exp(-1j * np.pi / 4)]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Ti._matrix

    @property
    def gate_type(self) -> str:
        return "ti"


@_from_braket_instruction.register(braket_instruction.Ti)
def _ti(instruction) -> Ti:
    return Ti([instruction.target])


class V(GateOperation):
    """Square root of the X (not) gate"""

    _matrix = np.array([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return V._matrix

    @property
    def gate_type(self) -> str:
        return "v"


@_from_braket_instruction.register(braket_instruction.V)
def _v(instruction) -> V:
    return V([instruction.target])


class Vi(GateOperation):
    r"""The adjoint :math:`V^{\dagger}` of the square root of the X (not) gate"""

    _matrix = np.array(([[0.5 - 0.5j, 0.5 + 0.5j], [0.5 + 0.5j, 0.5 - 0.5j]]), dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Vi._matrix

    @property
    def gate_type(self) -> str:
        return "vi"


@_from_braket_instruction.register(braket_instruction.Vi)
def _vi(instruction) -> Vi:
    return Vi([instruction.target])


class PhaseShift(GateOperation):
    """Phase shift gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.array([[1, 0], [0, np.exp(1j * angle_float)]], dtype=complex)

    @property
    def gate_type(self) -> str:
        return "phaseshift"


@_from_braket_instruction.register(braket_instruction.PhaseShift)
def _phase_shift(instruction) -> PhaseShift:
    return PhaseShift([instruction.target], instruction.angle)


class CPhaseShift(GateOperation):
    """Controlled phase shift gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.diag([1.0, 1.0, 1.0, np.exp(1j * angle_float)]).astype(complex)

    @property
    def gate_type(self) -> str:
        return "cphaseshift"


@_from_braket_instruction.register(braket_instruction.CPhaseShift)
def _c_phase_shift(instruction) -> CPhaseShift:
    return CPhaseShift([instruction.control, instruction.target], instruction.angle)


class CPhaseShift00(GateOperation):
    r"""Controlled phase shift gate phasing the :math:`\ket{00}` state"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.diag([np.exp(1j * angle_float), 1.0, 1.0, 1.0]).astype(complex)

    @property
    def gate_type(self) -> str:
        return "cphaseshift00"


@_from_braket_instruction.register(braket_instruction.CPhaseShift00)
def _c_phase_shift_00(instruction) -> CPhaseShift00:
    return CPhaseShift00([instruction.control, instruction.target], instruction.angle)


class CPhaseShift01(GateOperation):
    r"""Controlled phase shift gate phasing the :math:`\ket{01}` state"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.diag([1.0, np.exp(1j * angle_float), 1.0, 1.0]).astype(complex)

    @property
    def gate_type(self) -> str:
        return "cphaseshift01"


@_from_braket_instruction.register(braket_instruction.CPhaseShift01)
def _c_phase_shift_01(instruction) -> CPhaseShift01:
    return CPhaseShift01([instruction.control, instruction.target], instruction.angle)


class CPhaseShift10(GateOperation):
    r"""Controlled phase shift gate phasing the :math:`\ket{10}` state"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.diag([1.0, 1.0, np.exp(1j * angle_float), 1.0]).astype(complex)

    @property
    def gate_type(self) -> str:
        return "cphaseshift10"


@_from_braket_instruction.register(braket_instruction.CPhaseShift10)
def _c_phase_shift_10(instruction) -> CPhaseShift10:
    return CPhaseShift10([instruction.control, instruction.target], instruction.angle)


class RotX(GateOperation):
    """X-axis rotation gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        cos_half_angle = np.cos(angle_float / 2).astype(complex)
        i_sin_half_angle = 1j * np.sin(angle_float / 2).astype(complex)
        return np.array(
            [[cos_half_angle, -i_sin_half_angle], [-i_sin_half_angle, cos_half_angle]],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "rx"


@_from_braket_instruction.register(braket_instruction.Rx)
def _rot_x(instruction) -> RotX:
    return RotX([instruction.target], instruction.angle)


class RotY(GateOperation):
    """Y-axis rotation gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        cos_half_angle = np.cos(angle_float / 2).astype(complex)
        sin_half_angle = np.sin(angle_float / 2).astype(complex)
        return np.array(
            [[cos_half_angle, -sin_half_angle], [sin_half_angle, cos_half_angle]], dtype=complex
        )

    @property
    def gate_type(self) -> str:
        return "ry"


@_from_braket_instruction.register(braket_instruction.Ry)
def _rot_y(instruction) -> RotY:
    return RotY([instruction.target], instruction.angle)


class RotZ(GateOperation):
    """Z-axis rotation gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        return np.array(
            [[np.exp(-1j * angle_float / 2), 0], [0, np.exp(1j * angle_float / 2)]], dtype=complex
        )

    @property
    def gate_type(self) -> str:
        return "rz"


@_from_braket_instruction.register(braket_instruction.Rz)
def _rot_z(instruction) -> RotZ:
    return RotZ([instruction.target], instruction.angle)


class Swap(GateOperation):
    """Swap gate"""

    _matrix = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=complex)

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return Swap._matrix

    @property
    def gate_type(self) -> str:
        return "swap"


@_from_braket_instruction.register(braket_instruction.Swap)
def _swap(instruction) -> Swap:
    return Swap(instruction.targets)


class ISwap(GateOperation):
    """iSwap gate"""

    _matrix = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0j, 0.0],
            [0.0, 1.0j, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype=complex,
    )

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return ISwap._matrix

    @property
    def gate_type(self) -> str:
        return "iswap"


@_from_braket_instruction.register(braket_instruction.ISwap)
def _iswap(instruction) -> ISwap:
    return ISwap(instruction.targets)


class PSwap(GateOperation):
    """Parametrized Swap gate"""

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        exp_angle = np.exp(1j * angle_float)
        return np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, exp_angle, 0.0],
                [0.0, exp_angle, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "pswap"


@_from_braket_instruction.register(braket_instruction.PSwap)
def _pswap(instruction) -> PSwap:
    return PSwap(instruction.targets, instruction.angle)


class XY(GateOperation):
    """XY gate

    Reference: https://arxiv.org/abs/1912.04424v1
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        cos = np.cos(angle_float / 2).astype(complex)
        sin = np.sin(angle_float / 2).astype(complex)
        return np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, cos, 1.0j * sin, 0.0],
                [0.0, 1.0j * sin, cos, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "xy"


@_from_braket_instruction.register(braket_instruction.XY)
def _xy(instruction) -> XY:
    return XY(instruction.targets, instruction.angle)


class XX(GateOperation):
    """Ising XX gate

    Reference: https://arxiv.org/abs/1707.06356
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        cos_angle = np.cos(angle_float / 2).astype(complex)
        i_sin_angle = 1j * np.sin(angle_float / 2).astype(complex)
        return np.array(
            [
                [cos_angle, 0, 0, -i_sin_angle],
                [0, cos_angle, -i_sin_angle, 0],
                [0, -i_sin_angle, cos_angle, 0],
                [-i_sin_angle, 0, 0, cos_angle],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "xx"


@_from_braket_instruction.register(braket_instruction.XX)
def _xx(instruction) -> XX:
    return XX(instruction.targets, instruction.angle)


class YY(GateOperation):
    """Ising YY gate

    Reference: https://arxiv.org/abs/1707.06356
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        cos_angle = np.cos(angle_float / 2).astype(complex)
        i_sin_angle = 1j * np.sin(angle_float / 2).astype(complex)
        return np.array(
            [
                [cos_angle, 0, 0, i_sin_angle],
                [0, cos_angle, -i_sin_angle, 0],
                [0, -i_sin_angle, cos_angle, 0],
                [i_sin_angle, 0, 0, cos_angle],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "yy"


@_from_braket_instruction.register(braket_instruction.YY)
def _yy(instruction) -> YY:
    return YY(instruction.targets, instruction.angle)


class ZZ(GateOperation):
    """Ising ZZ gate

    Reference: https://arxiv.org/abs/1707.06356
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        positive_phase = np.exp(1j * angle_float / 2)
        negative_phase = np.exp(-1j * angle_float / 2)
        return np.array(
            [
                [negative_phase, 0, 0, 0],
                [0, positive_phase, 0, 0],
                [0, 0, positive_phase, 0],
                [0, 0, 0, negative_phase],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "zz"


@_from_braket_instruction.register(braket_instruction.ZZ)
def _zz(instruction) -> ZZ:
    return ZZ(instruction.targets, instruction.angle)


class CCNot(GateOperation):
    """Controlled CNot or Toffoli gate"""

    _matrix = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 1, 0],
        ],
        dtype=complex,
    )

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return CCNot._matrix

    @property
    def gate_type(self) -> str:
        return "ccnot"


@_from_braket_instruction.register(braket_instruction.CCNot)
def _ccnot(instruction) -> CCNot:
    return CCNot([*instruction.controls, instruction.target])


class CSwap(GateOperation):
    """Controlled Swap gate"""

    _matrix = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
        ],
        dtype=complex,
    )

    def __init__(self, targets, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )

    @property
    def _base_matrix(self) -> np.ndarray:
        return CSwap._matrix

    @property
    def gate_type(self) -> str:
        return "cswap"


@_from_braket_instruction.register(braket_instruction.CSwap)
def _cswap(instruction) -> CSwap:
    return CSwap([instruction.control, *instruction.targets])


class PRx(GateOperation):
    """
    PhaseRx gate.
    """

    def __init__(self, targets, angle_1, angle_2, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle_1 = angle_1
        self._angle_2 = angle_2

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_1_float = float(self._angle_1)
        angle_2_float = float(self._angle_2)
        cos = np.cos(angle_1_float / 2)
        isin = -1j * np.sin(angle_1_float / 2)
        exp_neg_angle_2 = np.exp(-1j * angle_2_float)
        exp_pos_angle_2 = np.exp(1j * angle_2_float)
        return np.array(
            [
                [
                    cos,
                    exp_neg_angle_2 * isin,
                ],
                [
                    exp_pos_angle_2 * isin,
                    cos,
                ],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "prx"


class GPi(GateOperation):
    """
    IonQ GPi gate.

    Reference: https://ionq.com/docs/getting-started-with-native-gates#single-qubit-gates
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        exp_neg_angle = np.exp(-1j * angle_float)
        exp_pos_angle = np.exp(1j * angle_float)
        return np.array(
            [
                [0, exp_neg_angle],
                [exp_pos_angle, 0],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "gpi"


class GPi2(GateOperation):
    """
    IonQ GPi2 gate.

    Reference: https://ionq.com/docs/getting-started-with-native-gates#single-qubit-gates
    """

    def __init__(self, targets, angle, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle = angle

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_float = float(self._angle)
        exp_neg_angle = np.exp(-1j * angle_float)
        exp_pos_angle = np.exp(1j * angle_float)
        return np.array(
            [
                [1, -1j * exp_neg_angle],
                [-1j * exp_pos_angle, 1],
            ],
            dtype=complex,
        ) / np.sqrt(2)

    @property
    def gate_type(self) -> str:
        return "gpi2"


class MS(GateOperation):
    """
    IonQ MS gate.

    Reference: https://ionq.com/docs/getting-started-with-native-gates#entangling-gates
    """

    def __init__(self, targets, angle_1, angle_2, angle_3=np.pi / 2, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._angle_1 = angle_1
        self._angle_2 = angle_2
        self._angle_3 = angle_3

    @property
    def _base_matrix(self) -> np.ndarray:
        angle_1_float = float(self._angle_1)
        angle_2_float = float(self._angle_2)
        angle_3_float = float(self._angle_3)
        cos = np.cos(angle_3_float / 2)
        isin = -1j * np.sin(angle_3_float / 2)
        exp_neg_sum = np.exp(-1j * (angle_1_float + angle_2_float))
        exp_neg_diff = np.exp(-1j * (angle_1_float - angle_2_float))
        exp_pos_diff = np.exp(1j * (angle_1_float - angle_2_float))
        exp_pos_sum = np.exp(1j * (angle_1_float + angle_2_float))
        return np.array(
            [
                [
                    cos,
                    0,
                    0,
                    exp_neg_sum * isin,
                ],
                [
                    0,
                    cos,
                    exp_neg_diff * isin,
                    0,
                ],
                [
                    0,
                    exp_pos_diff * isin,
                    cos,
                    0,
                ],
                [
                    exp_pos_sum * isin,
                    0,
                    0,
                    cos,
                ],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "ms"


class Unitary(GateOperation):
    """Arbitrary unitary gate"""

    def __init__(self, targets, matrix, ctrl_modifiers=(), power=1):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        clone = np.array(matrix, dtype=complex)
        check_matrix_dimensions(clone, self._targets)
        check_unitary(clone)
        self._matrix = clone

    @property
    def _base_matrix(self) -> np.ndarray:
        return self._matrix

    @property
    def gate_type(self) -> str:
        return "unitary"


@_from_braket_instruction.register(braket_instruction.Unitary)
def _unitary(instruction) -> Unitary:
    return Unitary(instruction.targets, ir_matrix_to_ndarray(instruction.matrix))


"""
OpenQASM gate operations
"""


class U(GateOperation):
    """
    Parameterized primitive gate for OpenQASM simulator
    """

    def __init__(
        self,
        targets: Sequence[int],
        theta: float,
        phi: float,
        lambda_: float,
        ctrl_modifiers: Sequence[int] = (),
        power: float = 1,
    ):
        super().__init__(
            targets=targets,
            ctrl_modifiers=ctrl_modifiers,
            power=power,
        )
        self._theta = theta
        self._phi = phi
        self._lambda = lambda_

    @property
    def _base_matrix(self) -> np.ndarray:
        """
        Generate parameterized Unitary matrix.
        https://openqasm.com/language/gates.html#built-in-gates

        Returns:
            np.ndarray: U Matrix
        """
        theta_float = float(self._theta)
        phi_float = float(self._phi)
        lambda_float = float(self._lambda)
        cos = np.cos(theta_float / 2).astype(complex)
        sin = np.sin(theta_float / 2).astype(complex)
        exp_lambda = np.exp(1j * lambda_float)
        exp_phi = np.exp(1j * phi_float)
        exp_phi_lambda = np.exp(1j * (phi_float + lambda_float))
        return np.array(
            [
                [cos, -exp_lambda * sin],
                [exp_phi * sin, exp_phi_lambda * cos],
            ],
            dtype=complex,
        )

    @property
    def gate_type(self) -> str:
        return "u"


class GPhase(GateOperation):
    """
    Global phase operation for OpenQASM simulator
    """

    def __init__(self, targets: Sequence[int], angle: float):
        super().__init__(
            targets=targets,
        )
        angle_float = float(angle)
        self._exp = np.exp(angle_float * 1j) * np.eye(2 ** len(self._targets), dtype=complex)

    @property
    def _base_matrix(self) -> np.ndarray:
        return self._exp

    @property
    def gate_type(self) -> str:
        return "gphase"


BRAKET_GATES = {
    "i": Identity,
    "h": Hadamard,
    "x": PauliX,
    "y": PauliY,
    "z": PauliZ,
    "cv": CV,
    "cnot": CX,
    "cy": CY,
    "cz": CZ,
    "ecr": ECR,
    "s": S,
    "si": Si,
    "t": T,
    "ti": Ti,
    "v": V,
    "vi": Vi,
    "phaseshift": PhaseShift,
    "cphaseshift": CPhaseShift,
    "cphaseshift00": CPhaseShift00,
    "cphaseshift01": CPhaseShift01,
    "cphaseshift10": CPhaseShift10,
    "rx": RotX,
    "ry": RotY,
    "rz": RotZ,
    "swap": Swap,
    "iswap": ISwap,
    "pswap": PSwap,
    "xy": XY,
    "xx": XX,
    "yy": YY,
    "zz": ZZ,
    "ccnot": CCNot,
    "cswap": CSwap,
    "prx": PRx,
    "gpi": GPi,
    "gpi2": GPi2,
    "ms": MS,
    "unitary": Unitary,
    "U": U,
    "gphase": GPhase,
}



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.analog_hamiltonian_simulator.rydberg.validators.atom_arrangement import (
    AtomArrangementValidator,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.capabilities_constants import (
    CapabilitiesConstants,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.driving_field import (
    DrivingFieldValidator,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.hamiltonian import HamiltonianValidator
from braket.analog_hamiltonian_simulator.rydberg.validators.local_detuning import (
    LocalDetuningValidator,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.physical_field import (
    PhysicalFieldValidator,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.program import ProgramValidator
from braket.analog_hamiltonian_simulator.rydberg.validators.times_series import TimeSeriesValidator
from braket.ir.ahs.program_v1 import Program


def validate_program(program: Program, device_capabilities: CapabilitiesConstants) -> None:
    """
    Validate the analog Hamiltonian simulation program has only one driving and shifting field,
    and all the sequences have the same last time point.

    Args:
        program (Program): An analog Hamiltonian simulation program
        device_capabilities (CapabilitiesConstants): The capability constants for the simulator
    """

    ProgramValidator(capabilities=device_capabilities, **program.dict())
    AtomArrangementValidator(capabilities=device_capabilities, **program.setup.ahs_register.dict())
    HamiltonianValidator(**program.hamiltonian.dict())
    for d_fields in program.hamiltonian.drivingFields:
        DrivingFieldValidator(capabilities=device_capabilities, **d_fields.dict())
        amplitude = d_fields.amplitude
        phase = d_fields.phase
        detuning = d_fields.detuning
        PhysicalFieldValidator(**amplitude.dict())
        TimeSeriesValidator(capabilities=device_capabilities, **amplitude.time_series.dict())

        PhysicalFieldValidator(**phase.dict())
        TimeSeriesValidator(capabilities=device_capabilities, **phase.time_series.dict())

        PhysicalFieldValidator(**detuning.dict())
        TimeSeriesValidator(capabilities=device_capabilities, **detuning.time_series.dict())
    for s_fields in program.hamiltonian.localDetuning:
        LocalDetuningValidator(capabilities=device_capabilities, **s_fields.dict())
        magnitude = s_fields.magnitude
        PhysicalFieldValidator(**magnitude.dict())
        TimeSeriesValidator(capabilities=device_capabilities, **magnitude.time_series.dict())



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import ABC, abstractmethod
from collections.abc import Sequence
from multiprocessing import Pool
from os import cpu_count
from typing import Optional, Union

from braket.device_schema import DeviceCapabilities
from braket.ir.ahs import Program as AHSProgram
from braket.ir.jaqcd import Program as JaqcdProgram
from braket.ir.openqasm import Program as OQ3Program
from braket.ir.openqasm.program_set_v1 import ProgramSet
from braket.task_result import AnalogHamiltonianSimulationTaskResult, GateModelTaskResult


class BraketSimulator(ABC):
    """An abstract simulator that locally runs a quantum task.

    The task can be either a quantum circuit defined in an OpenQASM or JAQCD program,
    or an analog Hamiltonian simulation (AHS) program.

    For users creating their own simulator: to register a simulator so the
    Braket SDK recognizes its name, the name and class must be added as an
    entry point for "braket.simulators". This is done by adding an entry to
    entry_points in the simulator package's setup.py:

    >>> entry_points = {
    >>>     "braket.simulators": [
    >>>         "backend_name = <backend_class>"
    >>>     ]
    >>> }
    """

    DEVICE_ID = None

    @abstractmethod
    def run(
        self, ir: Union[OQ3Program, AHSProgram, JaqcdProgram, ProgramSet], *args, **kwargs
    ) -> Union[GateModelTaskResult, AnalogHamiltonianSimulationTaskResult]:
        """
        Run the task specified by the given IR.

        Extra arguments will contain any additional information necessary to run the task,
        such as the extra parameters for AHS simulations.

        Args:
            ir (Union[OQ3Program, AHSProgram, JaqcdProgram]): The IR representation of the program

        Returns:
            Union[GateModelTaskResult, AnalogHamiltonianSimulationTaskResult]: An object
            representing the results of the simulation.
        """

    def run_multiple(
        self,
        programs: Sequence[Union[OQ3Program, AHSProgram, JaqcdProgram]],
        max_parallel: Optional[int] = None,
        *args,
        **kwargs,
    ) -> list[Union[GateModelTaskResult, AnalogHamiltonianSimulationTaskResult]]:
        """
        Run the tasks specified by the given IR programs.

        Extra arguments will contain any additional information necessary to run the tasks,
        such as the extra parameters for AHS simulations.

        Args:
            programs (Sequence[Union[OQ3Program, AHSProgram, JaqcdProgram]]): The IR representations
                of the programs
            max_parallel (Optional[int]): The maximum number of programs to run in parallel.
                Default is the number of logical CPUs.

        Returns:
            list[Union[GateModelTaskResult, AnalogHamiltonianSimulationTaskResult]]: A list of
            result objects, with the ith object being the result of the ith program.
        """
        max_parallel = max_parallel or cpu_count()
        with Pool(min(max_parallel, len(programs))) as pool:
            param_list = [(program, args, kwargs) for program in programs]
            results = pool.starmap(self._run_wrapped, param_list)
        return results

    def _run_wrapped(
        self, ir: Union[OQ3Program, AHSProgram, JaqcdProgram], args, kwargs
    ):  # pragma: no cover
        return self.run(ir, *args, **kwargs)

    @property
    @abstractmethod
    def properties(self) -> DeviceCapabilities:
        """DeviceCapabilities: Properties of the device."""



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from pydantic.v1.class_validators import root_validator

from braket.analog_hamiltonian_simulator.rydberg.validators.capabilities_constants import (
    CapabilitiesConstants,
)
from braket.analog_hamiltonian_simulator.rydberg.validators.field_validator_util import (
    validate_value_range_with_warning,
)
from braket.ir.ahs.local_detuning import LocalDetuning


class LocalDetuningValidator(LocalDetuning):
    capabilities: CapabilitiesConstants

    @root_validator(pre=True, skip_on_failure=True)
    def magnitude_pattern_is_not_uniform(cls, values):
        magnitude = values["magnitude"]
        pattern = magnitude["pattern"]
        if isinstance(pattern, str):
            raise ValueError(f"Pattern of shifting field must be not be a string - {pattern}")
        return values

    @root_validator(pre=True, skip_on_failure=True)
    def magnitude_pattern_within_bounds(cls, values):
        magnitude = values["magnitude"]
        capabilities = values["capabilities"]
        pattern = magnitude["pattern"]
        for index, pattern_value in enumerate(pattern):
            if (pattern_value < capabilities.MAGNITUDE_PATTERN_VALUE_MIN) or (
                pattern_value > capabilities.MAGNITUDE_PATTERN_VALUE_MAX
            ):
                raise ValueError(
                    f"magnitude pattern value {index} is {pattern_value}; it must be between "
                    f"{capabilities.MAGNITUDE_PATTERN_VALUE_MIN} and "
                    f"{capabilities.MAGNITUDE_PATTERN_VALUE_MAX} (inclusive)."
                )
        return values

    @root_validator(pre=True, skip_on_failure=True)
    def magnitude_values_within_range(cls, values):
        magnitude = values["magnitude"]
        capabilities = values["capabilities"]
        magnitude_values = magnitude["time_series"]["values"]
        validate_value_range_with_warning(
            magnitude_values,
            capabilities.LOCAL_MAGNITUDE_SEQUENCE_VALUE_MIN,
            capabilities.LOCAL_MAGNITUDE_SEQUENCE_VALUE_MAX,
            "magnitude",
        )
        return values



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import warnings

from braket.analog_hamiltonian_simulator.rydberg.constants import MIN_BLOCKADE_RADIUS


def validate_blockade_radius(blockade_radius: float) -> float:
    """Validate the Blockade radius

    Args:
        blockade_radius (float): The user-specified blockade radius

    Returns:
        float: The validated blockade radius

    Raises:
        ValueError: If blockade_radius < 0
    """
    blockade_radius = float(blockade_radius)
    if blockade_radius < 0:
        raise ValueError("`blockade_radius` needs to be non-negative.")

    if 0 < blockade_radius and blockade_radius < MIN_BLOCKADE_RADIUS:
        warnings.warn(
            f"Blockade radius {blockade_radius} meter is smaller than the typical value "
            f"({MIN_BLOCKADE_RADIUS} meter). "
            "The blockade radius should be specified in SI units."
        )

    return blockade_radius



# Apache V2 usage with no significant modifications
# Source: https://github.com/openqasm/openqasm/blob/429781bb9c95ef15944861f306ac6b9e4ff0abf0/source/openqasm/openqasm3/ast.py
# License: https://github.com/openqasm/openqasm/blob/main/LICENSE

"""
========================================
Abstract Syntax Tree (``openqasm3.ast``)
========================================

.. currentmodule:: openqasm3.ast

The reference abstract syntax tree (AST) for OpenQASM 3 programs.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

__all__ = [
    "AccessControl",
    "AliasStatement",
    "AngleType",
    "Annotation",
    "ArrayLiteral",
    "ArrayReferenceType",
    "ArrayType",
    "AssignmentOperator",
    "BinaryExpression",
    "BinaryOperator",
    "BitType",
    "BitstringLiteral",
    "BoolType",
    "BooleanLiteral",
    "Box",
    "BranchingStatement",
    "BreakStatement",
    "CalibrationDefinition",
    "CalibrationGrammarDeclaration",
    "Cast",
    "ClassicalArgument",
    "ClassicalAssignment",
    "ClassicalDeclaration",
    "ClassicalType",
    "ComplexType",
    "Concatenation",
    "ConstantDeclaration",
    "ContinueStatement",
    "DelayInstruction",
    "DiscreteSet",
    "DurationLiteral",
    "DurationOf",
    "DurationType",
    "EndStatement",
    "Expression",
    "ExpressionStatement",
    "ExternDeclaration",
    "FloatLiteral",
    "FloatType",
    "ForInLoop",
    "FunctionCall",
    "GateModifierName",
    "IODeclaration",
    "IOKeyword",
    "Identifier",
    "Include",
    "IndexExpression",
    "IndexedIdentifier",
    "IntType",
    "IntegerLiteral",
    "Pragma",
    "Program",
    "QASMNode",
    "QuantumArgument",
    "QuantumBarrier",
    "QuantumGate",
    "QuantumGateDefinition",
    "QuantumGateModifier",
    "QuantumMeasurement",
    "QuantumMeasurementStatement",
    "QuantumPhase",
    "QuantumReset",
    "QuantumStatement",
    "QubitDeclaration",
    "RangeDefinition",
    "ReturnStatement",
    "SizeOf",
    "Span",
    "Statement",
    "StretchType",
    "SubroutineDefinition",
    "TimeUnit",
    "UintType",
    "UnaryExpression",
    "UnaryOperator",
    "WhileLoop",
]

from sympy import Expr

AccessControl = Enum("AccessControl", "const mutable")
AssignmentOperator = Enum("AssignmentOperator", "= += -= *= /= &= |= ~= ^= <<= >>= %= **=")
BinaryOperator = Enum("BinaryOperator", "> < >= <= == != && || | ^ & << >> + - * / % **")
GateModifierName = Enum("GateModifier", "inv pow ctrl negctrl")
IOKeyword = Enum("IOKeyword", "input output")
TimeUnit = Enum("TimeUnit", "dt ns us ms s")
UnaryOperator = Enum("UnaryOperator", "~ ! -")


@dataclass
class Span:
    """
    Start and end line/column in the source file
    We use the Antlr convention. The starting line number is 1 and starting column number is 0.
    """

    start_line: int
    start_column: int
    end_line: int
    end_column: int


@dataclass
class QASMNode:
    """Base class for all OpenQASM 3 nodes"""

    span: Span | None = field(init=False, default=None, compare=False)
    """
    The span(location) of the node in the source code.
    Because not all the nodes are generated from source, the span is optional.
    To make it easier to write unit test, we exclude span from the generated __eq__().
    """


@dataclass
class Program(QASMNode):
    """
    An entire OpenQASM 3 program represented by a list of top level statements
    """

    statements: list[Statement]
    version: str | None = None


@dataclass
class Annotation(QASMNode):
    """An annotation applied to a statment."""

    keyword: str
    command: str | None = None


@dataclass
class Statement(QASMNode):
    """A statement: anything that can appear on its own line"""

    annotations: list[Annotation] = field(init=False, default_factory=list)


@dataclass
class Include(Statement):
    """
    An include statement
    """

    filename: str


@dataclass
class ExpressionStatement(Statement):
    """A statement that contains a single expression"""

    expression: Expression


# Note that QubitDeclaration is not a valid QuantumStatement, because qubits
# can only be declared in global scopes, not in gates.
@dataclass
class QubitDeclaration(Statement):
    """
    Global qubit declaration

    Example::

        qubit q;
        qubit[4] q;

        q // <- qubit
        4 // <- size

    """

    qubit: Identifier
    size: Expression | None = None


@dataclass
class QuantumGateDefinition(Statement):
    """
    Define a new quantum gate

    Example::

        gate cx c, t {
            ctrl @ unitary(pi, 0, pi) c, t;
        }

    """

    name: Identifier
    arguments: list[Identifier]
    qubits: list[Identifier]
    body: list[QuantumStatement]


class QuantumStatement(Statement):
    """Statements that may appear inside a gate declaration"""


@dataclass
class ExternDeclaration(Statement):
    """
    A extern declaration

    Example::

        extern get_pauli(int[prec]) -> bit[2 * n];

        get_pauli  // <- name
        int[prec]  // <- classical type
        bit[2 * n] // <- return type

    """

    name: Identifier
    arguments: list[ExternArgument]
    return_type: ExternArgument | None = None


class Expression(QASMNode):
    """An expression: anything that returns a value"""


@dataclass
class Identifier(Expression):
    """
    An identifier

    Example::

        q1

    """

    name: str


@dataclass
class UnaryExpression(Expression):
    """
    A unary expression

    Example::

        ~b
        !bool
        -i

    """

    op: UnaryOperator
    expression: Expression


@dataclass
class BinaryExpression(Expression):
    """
    A binary expression

    Example::

        q1 || q2

    """

    op: BinaryOperator
    lhs: Expression
    rhs: Expression


@dataclass
class IntegerLiteral(Expression):
    """
    An integer literal

    Example::

        1

    """

    value: int


@dataclass
class FloatLiteral(Expression):
    """
    An real number literal

    Example::

        1.1

    """

    value: float


@dataclass
class BooleanLiteral(Expression):
    """
    A boolean expression

    Example::

        true
        false

    """

    value: bool


@dataclass
class BitstringLiteral(Expression):
    """A literal bitstring value.  The ``value`` is the numerical value of the
    bitstring, and the ``width`` is the number of digits given."""

    value: int
    width: int


@dataclass
class DurationLiteral(Expression):
    """
    A duration literal

    Example::

        1.0ns

    """

    value: float
    unit: TimeUnit


@dataclass
class ArrayLiteral(Expression):
    """Array literal, used to initialise declared arrays.

    For example::

        array[uint[8], 2] row = {1, 2};
        array[uint[8], 2, 2] my_array = {{1, 2}, {3, 4}};
        array[uint[8], 2, 2] my_array = {row, row};
    """

    values: list[Expression]


@dataclass
class FunctionCall(Expression):
    """
    A function call expression

    Example::

        foo(1)

        foo // <- name

    """

    name: Identifier
    arguments: list[Expression]


@dataclass
class Cast(Expression):
    """
    A cast call expression

    Example::

        counts += int[1](b);

    """

    type: ClassicalType
    argument: Expression


@dataclass
class DiscreteSet(QASMNode):
    """
    A set of discrete values.  This can be used for the values in a ``for``
    loop, or to index certain values out of a register::

        for i in {1, 2, 3} {}
        let alias = qubits[{2, 3, 4}];
    """

    values: list[Expression]


@dataclass
class RangeDefinition(QASMNode):
    """
    Range definition.

    Example::

        1:2
        1:1:10
        :
    """

    start: Expression | None
    end: Expression | None
    step: Expression | None


IndexElement = DiscreteSet | list[Expression | RangeDefinition]


@dataclass
class IndexExpression(Expression):
    """
    An index expression.

    Example::

        q[1]
    """

    collection: Expression
    index: IndexElement


@dataclass
class IndexedIdentifier(QASMNode):
    """An indentifier with index operators, such that it can be used as an
    lvalue.  The list of indices is subsequent index brackets, so in::

        a[{1, 2, 3}][0:1, 0:1]

    the list of indices will have two elements.  The first will be a
    :class:`.DiscreteSet`, and the second will be a list of two
    :class:`.RangeDefinition`\\ s.
    """

    name: Identifier
    indices: list[IndexElement]


@dataclass
class Concatenation(Expression):
    """
    Concatenation of two registers, for example::

        a ++ b
        a[2:3] ++ a[0:1]
    """

    lhs: Expression
    rhs: Expression


@dataclass
class QuantumGate(QuantumStatement):
    """
    Invoking a quantum gate

    Example::
        cx[dur] 0, 1;

        or

        ctrl @ p(λ) a, b;

        ctrl @ // <- quantumGateModifier
        p // <- quantumGateName
        λ // <- argument
        a, b // <- qubit
    """

    modifiers: list[QuantumGateModifier]
    name: Identifier
    arguments: list[Expression]
    qubits: list[IndexedIdentifier | Identifier]
    duration: Expression | None = None


@dataclass
class QuantumGateModifier(QASMNode):
    """
    A quantum gate modifier

    Attributes:
        modifier: 'inv', 'pow', or 'ctrl'
        expression: only pow modifier has expression.

    Example::

        inv @
        pow(1/2)
        ctrl
    """

    modifier: GateModifierName
    argument: Expression | None = None


@dataclass
class QuantumPhase(QuantumStatement):
    """
    A quantum phase instruction

    Example::

        ctrl @ gphase(λ) a;

        ctrl @ // <- quantumGateModifier
        λ // <- argument
        a // <- qubit

    """

    modifiers: list[QuantumGateModifier]
    argument: Expression
    qubits: list[IndexedIdentifier | Identifier]


# Not a full expression because it can only be used in limited contexts.
@dataclass
class QuantumMeasurement(QASMNode):
    """
    A quantum measurement instruction

    Example::

        measure q;
    """

    qubit: IndexedIdentifier | Identifier


# Note that this is not a QuantumStatement because it involves access to
# classical bits.
@dataclass
class QuantumMeasurementStatement(Statement):
    """Stand-alone statement of a quantum measurement, potentially assigning the
    result to a classical variable.  This is not the only statement that
    `measure` can appear in (it can also be in classical declaration statements
    and returns)."""

    measure: QuantumMeasurement
    target: IndexedIdentifier | Identifier | None


@dataclass
class QuantumBarrier(QuantumStatement):
    """
    A quantum barrier instruction

    Example::

        barrier q;
    """

    qubits: list[Expression]


@dataclass
class QuantumReset(QuantumStatement):
    """
    A reset instruction.

    Example::

        reset q;
    """

    qubits: IndexedIdentifier | Identifier


@dataclass
class ClassicalArgument(QASMNode):
    """
    Classical argument for a gate or subroutine declaration
    """

    type: ClassicalType
    name: Identifier
    access: AccessControl | None = None


@dataclass
class ExternArgument(QASMNode):
    """Classical argument for an extern declaration."""

    type: ClassicalType
    access: AccessControl | None = None


@dataclass
class ClassicalDeclaration(Statement):
    """
    Classical variable declaration

    Example::

        bit c;
    """

    type: ClassicalType
    identifier: Identifier
    init_expression: Expression | QuantumMeasurement | None = None


@dataclass
class IODeclaration(Statement):
    """
    Input/output variable declaration

    Exampe::

        input angle[16] theta;
        output bit select;
    """

    io_identifier: IOKeyword
    type: ClassicalType
    identifier: Identifier


@dataclass
class ConstantDeclaration(Statement):
    """
    Constant declaration

    Example::

        const int[16] n = 10;
    """

    type: ClassicalType
    identifier: Identifier
    init_expression: Expression


class ClassicalType(QASMNode):
    """
    Base class for classical type
    """


@dataclass
class IntType(ClassicalType):
    """
    Node representing a classical ``int`` (signed integer) type, with an
    optional precision.

    Example:

        int[8]
        int[16]
    """

    size: Expression | None = None


@dataclass
class UintType(ClassicalType):
    """
    Node representing a classical ``uint`` (unsigned integer) type, with an
    optional precision.

    Example:

        uint[8]
        uint[16]
    """

    size: Expression | None = None


@dataclass
class FloatType(ClassicalType):
    """
    Node representing the classical ``float`` type, with the particular IEEE-754
    floating-point size optionally specified.

    Example:

        float[16]
        float[64]
    """

    size: Expression | None = None


@dataclass
class ComplexType(ClassicalType):
    """
    Complex ClassicalType. Its real and imaginary parts are based on other classical types.

    Example::

        complex[float]
        complex[float[32]]
    """

    base_type: FloatType | None


@dataclass
class AngleType(ClassicalType):
    """
    Node representing the classical ``angle`` type, with an optional precision.

    Example::

        angle[8]
        angle[16]
    """

    size: Expression | None = None


@dataclass
class BitType(ClassicalType):
    """
    Node representing the classical ``bit`` type, with an optional size.

    Example::

        bit[8]
        creg[8]
    """

    size: Expression | None = None


class BoolType(ClassicalType):
    """
    Leaf node representing the Boolean classical type.
    """


@dataclass
class ArrayType(ClassicalType):
    """Type of arrays that include allocation of the storage.

    This is generally any array declared as a standard statement, but not
    arrays declared by being arguments to subroutines.
    """

    base_type: IntType | UintType | FloatType | AngleType | BitType | BoolType | ComplexType
    dimensions: list[Expression]


@dataclass
class ArrayReferenceType(ClassicalType):
    """Type of arrays that are a reference to an array with allocated storage.

    This is generally any array declared as a subroutine argument.  The
    dimensions can be either a list of expressions (one for each dimension), or
    a single expression, which is the number of dimensions.

    For example::

        // `a` will have dimensions `[IntegerLiteral(2)]` (with a list), because
        // it is a 1D array, with a length of 2.
        def f(const array[uint[8], 2] a) {}
        // `b` will have dimension `IntegerLiteral(3)` (no list), because it is
        // a 3D array, but we don't know the lengths of its dimensions.
        def f(const array[uint[8], #dim=3] b) {}
    """

    base_type: IntType | UintType | FloatType | AngleType | BitType | BoolType | ComplexType
    dimensions: Expression | list[Expression]


class DurationType(ClassicalType):
    """
    Leaf node representing the ``duration`` type.
    """


class StretchType(ClassicalType):
    """
    Leaf node representing the ``stretch`` type.
    """


@dataclass
class CalibrationGrammarDeclaration(Statement):
    """
    Calibration grammar declaration

    Example::

        defcalgrammar "openpulse";
    """

    name: str


@dataclass
class CalibrationDefinition(Statement):
    """
    Calibration definition

    Example::

        defcal rz(angle[20] theta) q {
            shift_phase drive(q), -theta;
        }
    """

    name: Identifier
    arguments: list[ClassicalArgument]
    qubits: list[Identifier]
    return_type: ClassicalType | None
    body: str


@dataclass
class SubroutineDefinition(Statement):
    """
    Subroutine definition

    Example::

        def measure(qubit q) -> bit {
            s q;
            h q;
            return measure q;
        }
    """

    name: Identifier
    arguments: list[ClassicalArgument | QuantumArgument]
    body: list[Statement]
    return_type: ClassicalType | None = None


@dataclass
class QuantumArgument(QASMNode):
    """
    Quantum argument for a subroutine declaration
    """

    name: Identifier
    size: Expression | None = None


@dataclass
class ReturnStatement(Statement):
    """
    Classical or quantum return statement

    Example::

        return measure q;

        return a + b

    """

    expression: Expression | QuantumMeasurement | None = None


class BreakStatement(Statement):
    """
    Break statement

    Example::

        break;
    """


class ContinueStatement(Statement):
    """
    Continue statement

    Example::

        continue;
    """


class EndStatement(Statement):
    """
    End statement

    Example::

        end;
    """


@dataclass
class BranchingStatement(Statement):
    """
    Branch (``if``) statement

    Example::

        if (temp == 1) {
            ry(-pi / 2) scratch[0];
        } else continue;
    """

    condition: Expression
    if_block: list[Statement]
    else_block: list[Statement]


@dataclass
class WhileLoop(Statement):
    """
    While loop

    Example::

        while(~success) {
            reset magic;
            ry(pi / 4) magic;
            success = distill(magic, scratch);
        }
    """

    while_condition: Expression
    block: list[Statement]


@dataclass
class ForInLoop(Statement):
    """
    For in loop

    Example::

        for i in [0: 2] {
            majority a[i], b[i + 1], a[i + 1];
        }
    """

    type: ClassicalType
    identifier: Identifier
    set_declaration: RangeDefinition | DiscreteSet | Identifier
    block: list[Statement]


@dataclass
class DelayInstruction(QuantumStatement):
    """
    Delay instruction

    Example::

        delay[start_stretch] $0;
    """

    duration: Expression
    qubits: list[IndexedIdentifier | Identifier]


@dataclass
class Box(QuantumStatement):
    """
    Timing box

    Example::

        box [maxdur] {
            delay[start_stretch] $0;
            x $0;
        }
    """

    duration: Expression | None
    body: list[QuantumStatement]


@dataclass
class DurationOf(Expression):
    """
    Duration Of

    Example::

        durationof({x $0;})
    """

    target: list[Statement]


@dataclass
class SizeOf(Expression):
    """``sizeof`` an array's dimensions."""

    target: Expression
    index: Expression | None = None


@dataclass
class AliasStatement(Statement):
    """
    Alias statement

    Example::

        let a = qubits[0];

    """

    target: Identifier
    value: Identifier | Concatenation


@dataclass
class ClassicalAssignment(Statement):
    """
    Classical assignment

    Example::

        a[0] = 1;
    """

    lvalue: Identifier | IndexedIdentifier
    op: AssignmentOperator
    rvalue: Expression


@dataclass
class Pragma(QASMNode):
    """
    Pragma
    Example::

        #pragma val1 val2 val3
    """

    command: str


@dataclass
class SymbolLiteral(Expression):
    """
    A symbol literal for parametric interpretation

    Example::

        Symbol("theta")

    """

    value: Expr



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import warnings

from pydantic.v1.class_validators import root_validator

from braket.analog_hamiltonian_simulator.rydberg.validators.capabilities_constants import (
    CapabilitiesConstants,
)
from braket.ir.ahs.time_series import TimeSeries


class TimeSeriesValidator(TimeSeries):
    capabilities: CapabilitiesConstants

    # must have at least 2 time values
    @root_validator(pre=True, skip_on_failure=True)
    def at_least_2_timepoints(cls, values):
        times = values["times"]
        length = len(times)
        min_length = 2
        if length < min_length:
            raise ValueError(f"Length of times must be at least {min_length}; it is {length}")
        return values

    # the first times entry should be 0
    @root_validator(pre=True, skip_on_failure=True)
    def times_start_with_0(cls, values):
        times = values["times"]
        if times[0] != 0.0:
            raise ValueError(f"First time value is {times[0]}; it must be 0.0")
        return values

    # The duration of the program should be below MAX_TIME
    # If not, a warning message will be issue to remind the user that the SI units are used here.
    @root_validator(pre=True, skip_on_failure=True)
    def times_are_not_too_big(cls, values):
        times = values["times"]
        capabilities = values["capabilities"]
        if times[-1] > capabilities.MAX_TIME:
            warnings.warn(
                f"Max time is {times[-1]} seconds which is bigger than the typical scale "
                f"({capabilities.MAX_TIME} seconds). "
                "The time points should be specified in SI units."
            )
        return values

    # The time array must be sorted in ascending order
    @root_validator(pre=True, skip_on_failure=True)
    def times_must_be_ascendingly_sorted(cls, values):
        times = values["times"]
        for i in range(len(times) - 1):
            if times[i] >= times[i + 1]:
                raise ValueError(
                    f"Time point {i} ({times[i]}) and time point {i + 1} ({times[i + 1]}) "
                    "must be monotonically increasing."
                )
        return values

    # Check that the times and the values have the same length
    @root_validator(pre=True, skip_on_failure=True)
    def check_times_and_values_have_same_length(cls, values):
        len_times = len(values["times"])
        len_values = len(values["values"])
        if len_values != len_times:
            raise ValueError(
                f"The sample times (length: {len_times}) and the values (length: {len_values}) "
                "must have the same length."
            )
        return values



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import itertools
import time
import warnings

import numpy as np
import scipy.sparse

from braket.ir.ahs.atom_arrangement import AtomArrangement
from braket.ir.ahs.program_v1 import Program


def validate_config(config: str, atoms_coordinates: np.ndarray, blockade_radius: float) -> bool:
    """Valid if a given configuration complies with the Rydberg approximation

    Args:
        config (str): The configuration to be validated
        atoms_coordinates (ndarray): The coordinates for atoms in the filled sites
        blockade_radius (float): The Rydberg blockade radius

    Returns:
        bool: True if the configuration complies with the Rydberg approximation,
        False otherwise
    """

    # The indices for the Rydberg atoms in the configuration
    rydberg_atoms = [i for i, item in enumerate(config) if item == "r"]

    for i, rydberg_atom in enumerate(rydberg_atoms[:-1]):
        dists = np.linalg.norm(
            atoms_coordinates[rydberg_atom] - atoms_coordinates[rydberg_atoms[i + 1 :]], axis=1
        )
        if min(dists) <= blockade_radius:
            return False
    return True


def get_blockade_configurations(lattice: AtomArrangement, blockade_radius: float) -> list[str]:
    """Return the lattice configurations complying with the blockade approximation

    Args:
        lattice (AtomArrangement): A lattice with Rydberg atoms and their coordinates
        blockade_radius (float): The Rydberg blockade radius

    Returns:
        list[str]: A list of bit strings, each of them corresponding to a valid
        configuration complying with the blockade approximation. The length of
        each configuration is the same as the number of atoms in the lattice,
        with 'r' and 'g' indicating the Rydberg and ground states, respectively.

        Notes on the indexing: The left-most bit in the configuration corresponds to
        the first atom in the lattice.

        Notes on the algorithm: We start from all possible configurations and get rid of
        those violating the blockade approximation constraint.
    """

    # The coordinates for atoms in the filled sites
    atoms_coordinates = np.array(lattice.sites)[np.where(lattice.filling)]
    min_separation = float("inf")  # The minimum separation between atoms, or filled sites
    for i, atom_coord in enumerate(atoms_coordinates[:-1]):
        dists = np.linalg.norm(atom_coord - atoms_coordinates[i + 1 :], axis=1)
        min_separation = min(min_separation, min(dists))

    configurations = [
        "".join(item) for item in itertools.product(["g", "r"], repeat=sum(lattice.filling))
    ]

    if blockade_radius < min_separation:  # no need to consider blockade approximation
        return configurations
    return [
        config
        for config in configurations
        if validate_config(config, atoms_coordinates, blockade_radius)
    ]


def _get_interaction_dict(
    program: Program, rydberg_interaction_coef: float, configurations: list[str]
) -> dict[tuple[int, int], float]:
    """Return the dict contains the Rydberg interaction strength for all configurations.

    Args:
        program (Program): An analog simulation program for Rydberg system with the interaction term
        rydberg_interaction_coef (float): The interaction coefficient
        configurations (list[str]): The list of configurations that comply with the blockade
            approximation.

    Returns:
        dict[tuple[int, int], float]: The dictionary for the interaction operator
    """

    # The coordinates for atoms in the filled sites
    lattice = program.setup.ahs_register
    atoms_coordinates = np.array(
        [lattice.sites[i] for i in range(len(lattice.sites)) if lattice.filling[i] == 1]
    )

    interactions = {}  # The interaction in the basis of configurations, as a dictionary

    for config_index, config in enumerate(configurations):
        interaction = 0

        # The indices for the Rydberg atoms in the configuration
        rydberg_atoms = [i for i, item in enumerate(config) if item == "r"]

        # Obtain the pairwise distances between the Rydberg atoms, followed by adding their Rydberg
        # interactions
        for ind_1, rydberg_atom_1 in enumerate(rydberg_atoms[:-1]):
            for ind_2, rydberg_atom_2 in enumerate(rydberg_atoms):
                if ind_2 > ind_1:
                    dist = np.linalg.norm(
                        atoms_coordinates[rydberg_atom_1] - atoms_coordinates[rydberg_atom_2]
                    )
                    interaction += rydberg_interaction_coef / (float(dist) ** 6)

        if interaction > 0:
            interactions[(config_index, config_index)] = interaction

    return interactions


def _get_detuning_dict(
    targets: tuple[int], configurations: list[str]
) -> dict[tuple[int, int], float]:
    """Return the dict contains the detuning operators for a set of target atoms.

    Args:
        targets (tuple[int]): The target atoms of the detuning operator
        configurations (list[str]): The list of configurations that comply with the blockade
            approximation.

    Returns:
        dict[tuple[int, int], float]: The dictionary for the detuning operator
    """

    detuning = {}  # The detuning term in the basis of configurations, as a dictionary

    for ind_1, config in enumerate(configurations):
        value = sum([1 for ind_2, item in enumerate(config) if item == "r" and ind_2 in targets])
        if value > 0:
            detuning[(ind_1, ind_1)] = value

    return detuning


def _get_rabi_dict(targets: tuple[int], configurations: list[str]) -> dict[tuple[int, int], float]:
    """Return the dict for the Rabi operators for a set of target atoms.

    Args:
        targets (tuple[int]): The target atoms of the detuning operator
        configurations (list[str]): The list of configurations that comply with the blockade
            approximation.

    Returns:
        dict[tuple[int, int], float]: The dictionary for the Rabi operator

    Note:
        We only save the lower triangular part of the matrix that corresponds
        to the Rabi operator.
    """

    rabi = {}  # The Rabi term in the basis of configurations, as a dictionary

    # use dictionary to store index of configurations
    configuration_index = {config: ind for ind, config in enumerate(configurations)}

    for ind_1, config_1 in enumerate(configurations):
        for target in targets:
            # Only keep the lower triangular part of the Rabi operator
            # which convert a single atom from "g" to "r".
            if config_1[target] != "g":
                continue

            # Construct the state after applying the Rabi operator
            bit_list = list(config_1)
            bit_list[target] = "r"
            config_2 = "".join(bit_list)

            # If the constructed state is in the Hilbert space,
            # add the corresponding matrix element to the Rabi operator.
            if config_2 in configuration_index:
                rabi[(configuration_index[config_2], ind_1)] = 1

    return rabi


def _get_sparse_from_dict(
    matrix_dict: dict[tuple[int, int], float], matrix_dimension: int
) -> scipy.sparse.csr_matrix:
    """Convert a dict to a CSR sparse matrix

    Args:
        matrix_dict (dict[tuple[int, int], float]): The dict for the sparse matrix
        matrix_dimension (int): The size of the sparse matrix

    Returns:
        scipy.sparse.csr_matrix: The sparse matrix in CSR format
    """
    rows = [key[0] for key in matrix_dict.keys()]
    cols = [key[1] for key in matrix_dict.keys()]
    return scipy.sparse.csr_matrix(
        tuple([list(matrix_dict.values()), [rows, cols]]),
        shape=(matrix_dimension, matrix_dimension),
    )


def _get_sparse_ops(
    program: Program, configurations: list[str], rydberg_interaction_coef: float
) -> tuple[
    list[scipy.sparse.csr_matrix],
    list[scipy.sparse.csr_matrix],
    scipy.sparse.csr_matrix,
    list[scipy.sparse.csr_matrix],
]:
    """Returns the sparse matrices for Rabi, detuning, interaction and local detuning detuning
    operators

    Args:
        program (Program): An analog simulation program for Rydberg system
        configurations (list[str]): The list of configurations that comply with the blockade
            approximation.
        rydberg_interaction_coef (float): The interaction coefficient

    Returns:
        tuple[list[csr_matrix],list[csr_matrix],csr_matrix,list[csr_matrix]]: A tuple containing
        the list of Rabi operators, the list of detuning operators,
        the interaction operator and the list of local detuning operators

    """
    # Get the driving fields as sparse matrices, whose targets are all the atoms in the system
    targets = np.arange(np.count_nonzero(program.setup.ahs_register.filling))
    rabi_dict = _get_rabi_dict(targets, configurations)
    detuning_dict = _get_detuning_dict(targets, configurations)

    # Driving field is an array of operators, which has only one element for now
    rabi_ops = [_get_sparse_from_dict(rabi_dict, len(configurations))]
    detuning_ops = [_get_sparse_from_dict(detuning_dict, len(configurations))]

    # Get the interaction term as a sparse matrix
    interaction_dict = _get_interaction_dict(program, rydberg_interaction_coef, configurations)
    interaction_op = _get_sparse_from_dict(interaction_dict, len(configurations))

    # Get local detuning as sparse matrices.
    # Local detuning is an array of operators, which has only one element for now
    local_detuning_ops = []
    for local_detuning in program.hamiltonian.localDetuning:
        temp = 0
        filled_site = 0  # Index of the filled site
        for filling, strength in zip(
            program.setup.ahs_register.filling, local_detuning.magnitude.pattern
        ):
            # If the site is not filled, we move on to the next filled site
            if filling == 0:
                continue
            opt = _get_sparse_from_dict(
                _get_detuning_dict((filled_site,), configurations), len(configurations)
            )
            temp += float(strength) * scipy.sparse.csr_matrix(opt, dtype=float)
            filled_site += 1

        local_detuning_ops.append(temp)

    return rabi_ops, detuning_ops, interaction_op, local_detuning_ops


def _interpolate_time_series(
    t: float, times: list[float], values: list[float], method: str = "piecewise_linear"
) -> float:
    """Interpolates the value of a series of time-value pairs at the given time via linear
        interpolation.

    Args:
        t (float): The given time point
        times (list[float]): The list of time points
        values (list[float]): The list of values
        method (str): The method for interpolation, either "piecewise_linear" or
            "piecewise_constant." Default: "piecewise_linear"

    Returns:
        float: The interpolated value of the time series at t
    """

    times = [float(time) for time in times]
    values = [float(value) for value in values]

    if method == "piecewise_linear":
        return np.interp(t, times, values)
    elif method == "piecewise_constant":
        index = np.searchsorted(times, t, side="right") - 1
        return values[index]
    else:
        raise ValueError("`method` can only be `piecewise_linear` or `piecewise_constant`.")


def _get_coefs(
    program: Program, simulation_times: list[float]
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Returns the coefficients for the Rabi operators, detuning operators and local detuning
    operators for all the time points in the analog simulation program.

    Args:
        program (Program): An analog simulation program for Rydberg system
        simulation_times (list[float]): The list of time points

    Returns:
        tuple[ndarray, ndarray, ndarray]: A tuple containing
        the list of Rabi frequencies, the list of global detunings and
        the list of local detunings
    """
    rabi_coefs, detuning_coefs = [], []

    for driving_field in program.hamiltonian.drivingFields:
        amplitude = driving_field.amplitude.time_series
        phase = driving_field.phase.time_series
        detuning = driving_field.detuning.time_series

        # Get the Rabi part. We use the convention: Omega * exp(1j*phi) * |r><g| + h.c.
        rabi_coef = np.array(
            [
                _interpolate_time_series(
                    t, amplitude.times, amplitude.values, method="piecewise_linear"
                )
                * np.exp(
                    1j
                    * _interpolate_time_series(
                        t, phase.times, phase.values, method="piecewise_constant"
                    )
                )
                for t in simulation_times
            ],
            dtype=complex,
        )
        rabi_coefs.append(rabi_coef)

        # Get the detuning part
        detuning_coef = np.array(
            [
                _interpolate_time_series(
                    t, detuning.times, detuning.values, method="piecewise_linear"
                )
                for t in simulation_times
            ],
            dtype=complex,
        )
        detuning_coefs.append(detuning_coef)

    # add local detuning
    local_detuning_coefs = []
    for local_detuning in program.hamiltonian.localDetuning:
        magnitude = local_detuning.magnitude.time_series

        local_detuning_coef = np.array(
            [
                _interpolate_time_series(
                    t, magnitude.times, magnitude.values, method="piecewise_linear"
                )
                for t in simulation_times
            ],
            dtype=complex,
        )
        local_detuning_coefs.append(local_detuning_coef)

    return np.array(rabi_coefs), np.array(detuning_coefs), np.array(local_detuning_coefs)


def _get_ops_coefs(
    program: Program,
    configurations: list[str],
    rydberg_interaction_coef: float,
    simulation_times: list[float],
) -> tuple[
    list[scipy.sparse.csr_matrix],
    list[scipy.sparse.csr_matrix],
    list[scipy.sparse.csr_matrix],
    np.ndarray,
    np.ndarray,
    np.ndarray,
    scipy.sparse.csr_matrix,
]:
    """Returns the sparse matrices and coefficients for the Rabi terms, detuning terms and
    the local detuning terms, together with the interaction operator in the given analog
    simulation program for Rydberg systems.

    Args:
        program (Program): An analog simulation program for Rydberg system
        configurations (list[str]): The list of configurations that comply to the
            blockade approximation.
        rydberg_interaction_coef (float): The interaction coefficient
        simulation_times (list[float]): The list of time points

    Returns:
        tuple[
            list[csr_matrix],
            list[csr_matrix],
            list[csr_matrix],
            ndarray,
            ndarray,
            ndarray,
            csr_matrix
        ]: A tuple containing the list of Rabi operators, the list of detuning operators,
        the list of local detuning operators, the list of Rabi frequencies, the list of global
        detunings, the list of local detunings and the interaction operator.
    """

    rabi_ops, detuning_ops, interaction_op, local_detuning_ops = _get_sparse_ops(
        program, configurations, rydberg_interaction_coef
    )
    rabi_coefs, detuning_coefs, local_detuning_coefs = _get_coefs(program, simulation_times)

    return (
        rabi_ops,
        detuning_ops,
        local_detuning_ops,
        rabi_coefs,
        detuning_coefs,
        local_detuning_coefs,
        interaction_op,
    )


def sample_state(state: np.ndarray, shots: int) -> np.ndarray:
    """Sample measurement outcomes from the quantum state `state`

    Args:
        state (ndarray): A state vector
        shots (int): The number of samples

    Returns:
        ndarray: The array for the sample results
    """

    weights = (np.abs(state) ** 2).flatten()
    weights /= sum(weights)
    sample = np.random.multinomial(shots, weights)
    return sample


def _print_progress_bar(num_time_points: int, index_time: int, start_time: float) -> None:
    """Print a lightweight progress bar

    Args:
        num_time_points (int): The total number of time points
        index_time (int): The index of the current time point
        start_time (float): The starting time for the simulation

    """
    if index_time == 0:
        print("0% finished, elapsed time = NA, ETA = NA", flush=True, end="\r")
    else:
        current_time = time.time()
        estimate_time_arrival = (
            (current_time - start_time) / (index_time + 1) * (num_time_points - (index_time + 1))
        )
        print(
            f"{100 * (index_time + 1) / num_time_points}% finished, "
            f"elapsed time = {(current_time - start_time)} seconds, "
            f"ETA = {estimate_time_arrival} seconds ",
            flush=True,
            end="\r",
        )


def _get_hamiltonian(
    index_time: float,
    operators_coefficients: tuple[
        list[scipy.sparse.csr_matrix],
        list[scipy.sparse.csr_matrix],
        list[scipy.sparse.csr_matrix],
        np.ndarray,
        np.ndarray,
        np.ndarray,
        scipy.sparse.csr_matrix,
    ],
) -> scipy.sparse.csr_matrix:
    """Get the Hamiltonian at a given time point

    Args:
        index_time (float): The index of the current time point
        operators_coefficients (tuple[
            list[csr_matrix],
            list[csr_matrix],
            list[csr_matrix],
            ndarray,
            ndarray,
            ndarray,
            csr_matrix
        ]): A tuple containing the list of Rabi operators, the list of detuning operators,
        the list of local detuning operators, the list of Rabi frequencies, the list of global
        detunings, the list of local detunings and the interaction operator.

    Returns:
        (scipy.sparse.csr_matrix): The Hamiltonian at the given time point as a sparse matrix
    """
    (
        rabi_ops,
        detuning_ops,
        local_detuning_ops,
        rabi_coefs,
        detuning_coefs,
        local_detuning_coefs,
        interaction_op,
    ) = operators_coefficients

    index_time = int(index_time)

    if len(rabi_coefs) > 0:
        # If there is driving field, the maximum of index_time is the maximum time index
        # for the driving field.
        # Note that, if there is more than one driving field, we assume that they have the
        # same number of coefficients
        max_index_time = len(rabi_coefs[0]) - 1
    else:
        # If there is no driving field, then the maxium of index_time is the maxium time
        # index for local detuning.
        # Note that, if there is more than one local detuning, we assume that they have the
        # same number of coefficients
        # Note that, if there is no driving field nor local detuning, the initial state will
        # be returned, and the simulation would not reach here.
        max_index_time = len(local_detuning_coefs[0]) - 1

    # If the integrator uses intermediate time value that is larger than the maximum
    # time value specified, the final time value is used as an approximation.
    if index_time > max_index_time:
        index_time = max_index_time
        warnings.warn(
            "The solver uses intermediate time value that is "
            "larger than the maximum time value specified. "
            "The final time value of the specified range "
            "is used as an approximation."
        )

    # If the integrator uses intermediate time value that is larger than the minimum
    # time value specified, the final time value is used as an approximation.
    if index_time < 0:
        index_time = 0
        warnings.warn(
            "The solver uses intermediate time value that is "
            "smaller than the minimum time value specified. "
            "The first time value of the specified range "
            "is used as an approximation."
        )

    hamiltonian = interaction_op

    # Add the driving fields
    for rabi_op, rabi_coef, detuning_op, detuning_coef in zip(
        rabi_ops, rabi_coefs, detuning_ops, detuning_coefs
    ):
        hamiltonian += (
            rabi_op * rabi_coef[index_time] / 2
            + (rabi_op.T.conj() * np.conj(rabi_coef[index_time]) / 2)
            - detuning_op * detuning_coef[index_time]
        )

    # Add local detuning
    for local_detuning_op, local_detuning_coef in zip(local_detuning_ops, local_detuning_coefs):
        hamiltonian -= local_detuning_op * local_detuning_coef[index_time]

    return hamiltonian


def _apply_hamiltonian(
    index_time: float,
    operators_coefficients: tuple[
        list[scipy.sparse.csr_matrix],
        list[scipy.sparse.csr_matrix],
        list[scipy.sparse.csr_matrix],
        np.ndarray,
        np.ndarray,
        np.ndarray,
        scipy.sparse.csr_matrix,
    ],
    input_register: np.ndarray,
) -> scipy.sparse.csr_matrix:
    """Applies the Hamiltonian at a given time point on a state.

    Args:
        index_time (float): The index of the current time point
        operators_coefficients (tuple[
            list[csr_matrix],
            list[csr_matrix],
            list[csr_matrix],
            ndarray,
            ndarray,
            ndarray,
            csr_matrix
        ]): A tuple containing the list of Rabi operators, the list of detuning operators,
        the list of local detuning operators, the list of Rabi frequencies, the list of global
        detunings, the list of local detunings and the interaction operator.
        input_register (ndarray): The input state which we apply the Hamiltonian to.
    Returns:
        (ndarray): The result
    """
    (
        rabi_ops,
        detuning_ops,
        local_detuning_ops,
        rabi_coefs,
        detuning_coefs,
        local_detuning_coefs,
        interaction_op,
    ) = operators_coefficients

    index_time = int(index_time)
    if len(rabi_coefs):
        # If there is driving field, the maximum of index_time is the maximum time index
        # for the driving field.
        # Note that, if there is more than one driving field, we assume that they have the
        # same number of coefficients
        max_index_time = len(rabi_coefs[0]) - 1
    else:
        # If there is no driving field, then the maxium of index_time is the maxium time
        # index for local detuning.
        # Note that, if there is more than one local detuning, we assume that they have the
        # same number of coefficients
        # Note that, if there is no driving field nor local detuning, the initial state will
        # be returned, and the simulation would not reach here.
        max_index_time = len(local_detuning_coefs[0]) - 1

    # If the integrator uses intermediate time value that is larger than the maximum
    # time value specified, the final time value is used as an approximation.
    if index_time > max_index_time:
        index_time = max_index_time
        warnings.warn(
            "The solver uses intermediate time value that is "
            "larger than the maximum time value specified. "
            "The final time value of the specified range "
            "is used as an approximation."
        )

    # If the integrator uses intermediate time value that is larger than the minimum
    # time value specified, the final time value is used as an approximation.
    if index_time < 0:
        index_time = 0
        warnings.warn(
            "The solver uses intermediate time value that is "
            "smaller than the minimum time value specified. "
            "The first time value of the specified range "
            "is used as an approximation."
        )

    output_register = interaction_op.dot(input_register)

    # Add the driving fields
    for rabi_op, rabi_coef, detuning_op, detuning_coef in zip(
        rabi_ops, rabi_coefs, detuning_ops, detuning_coefs
    ):
        output_register += (rabi_coef[index_time] / 2) * rabi_op.dot(input_register)
        output_register += (np.conj(rabi_coef[index_time]) / 2) * rabi_op.getH().dot(input_register)
        output_register -= detuning_coef[index_time] * detuning_op.dot(input_register)

    # Add local detuning
    for local_detuning_op, local_detuning_coef in zip(local_detuning_ops, local_detuning_coefs):
        output_register -= local_detuning_coef[index_time] * local_detuning_op.dot(input_register)

    return output_register



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np
from scipy.linalg import block_diag

from braket.default_simulator.linalg_utils import (
    QuantumGateDispatcher,
    multiply_matrix,
    partial_trace,
)
from braket.default_simulator.operation import GateOperation, KrausOperation, Observable
from braket.default_simulator.simulation import Simulation


class DensityMatrixSimulation(Simulation):
    """
    This class tracks the evolution of the density matrix of a quantum system with
    `qubit_count` qubits. The state of system evolves by applications of `GateOperation`s
    and `KrausOperation`s using the `evolve()` method.
    """

    def __init__(self, qubit_count: int, shots: int):
        """
        Args:
            qubit_count (int): The number of qubits being simulated.
            shots (int): The number of samples to take from the simulation.
                If set to 0, only results that do not require sampling, such as density matrix
                or expectation, are generated.
        """
        super().__init__(qubit_count=qubit_count, shots=shots)
        initial_state = np.zeros((2**qubit_count, 2**qubit_count), dtype=complex)
        initial_state[0, 0] = 1
        self._density_matrix = initial_state
        self._post_observables = None
        self._rng_generator = np.random.default_rng()

    def evolve(self, operations: list[GateOperation | KrausOperation]) -> None:
        self._density_matrix = DensityMatrixSimulation._apply_operations(
            self._density_matrix, self._qubit_count, operations
        )

    def apply_observables(self, observables: list[Observable]) -> None:
        """Applies the diagonalizing matrices of the given observables
        to the state of the simulation.

        This method can only be called once.

        Args:
            observables (list[Observable]): The observables to apply

        Raises:
            RuntimeError: If this method is called more than once
        """
        if self._post_observables is not None:
            raise RuntimeError("Observables have already been applied.")
        operations = [
            *sum(
                [observable.diagonalizing_gates(self._qubit_count) for observable in observables],
                (),
            )
        ]
        self._post_observables = DensityMatrixSimulation._apply_operations(
            self._density_matrix, self._qubit_count, operations
        )

    def retrieve_samples(self) -> np.ndarray:
        return np.searchsorted(
            np.cumsum(self.probabilities), self._rng_generator.random(size=self._shots)
        )

    @property
    def density_matrix(self) -> np.ndarray:
        """
        np.ndarray: The density matrix specifying the current state of the simulation.

        Note:
            Mutating this array will mutate the state of the simulation.
        """
        return self._density_matrix

    @property
    def state_with_observables(self) -> np.ndarray:
        """
        np.ndarray: The density matrix diagonalized in the basis of the measured observables.

        Raises:
            RuntimeError: If observables have not been applied
        """
        if self._post_observables is None:
            raise RuntimeError("No observables applied")
        return self._post_observables

    def expectation(self, observable: Observable) -> float:
        with_observables = observable.apply(
            np.reshape(self._density_matrix, [2] * 2 * self._qubit_count)
        )
        return complex(partial_trace(with_observables)).real

    @property
    def probabilities(self) -> np.ndarray:
        """
        np.ndarray: The probabilities of each computational basis state of the current density
            matrix of the simulation.
        """
        return DensityMatrixSimulation._probabilities(self.density_matrix)

    @staticmethod
    def _probabilities(state) -> np.ndarray:
        """The probabilities of each computational basis state of a given density matrix.

        Args:
            state (np.ndarray): The density matrix from which probabilities are extracted.

        Returns:
            np.ndarray: The probabilities of each computational basis state.
        """
        diag = np.real(np.diag(state))
        tol = 1e-20
        return np.where((np.abs(diag) >= tol) & (diag >= 0), diag, 0.0)

    @staticmethod
    def _apply_operations(
        state: np.ndarray,
        qubit_count: int,
        operations: list[GateOperation | KrausOperation | Observable],
    ) -> np.ndarray:
        """Applies the gate and noise operations to the density matrix.

        Args:
            state (np.ndarray): initial density matrix
            qubit_count (int): number of qubits in the circuit
            operations (list[Union[GateOperation, KrausOperation, Observable]]): list of GateOperation and
                KrausOperation to be applied to the density matrix

        Returns:
            np.ndarray: output density matrix
        """
        if not operations:
            return state
        dispatcher = QuantumGateDispatcher(state.size)
        original_shape = state.shape
        result = state.view()
        result.shape = [2] * 2 * qubit_count
        temp = np.zeros_like(result, dtype=complex)
        work_buffer1 = np.zeros_like(result, dtype=complex)
        work_buffer2 = np.zeros_like(result, dtype=complex)

        for operation in operations:
            if isinstance(operation, (GateOperation, Observable)):
                targets = operation.targets
                num_ctrl = len(operation._ctrl_modifiers)
                # Extract gate_type if available
                result, temp = DensityMatrixSimulation._apply_gate(
                    result,
                    temp,
                    qubit_count,
                    operation.matrix,
                    targets[num_ctrl:],
                    targets[:num_ctrl],
                    operation._ctrl_modifiers,
                    dispatcher,
                    getattr(operation, "gate_type"),
                )
            if isinstance(operation, KrausOperation):
                result, temp = DensityMatrixSimulation._apply_kraus(
                    result,
                    temp,
                    work_buffer1,
                    work_buffer2,
                    qubit_count,
                    operation.matrices,
                    operation.targets,
                    dispatcher,
                )
        result.shape = original_shape
        return result

    @staticmethod
    def _apply_gate(
        result: np.ndarray,
        temp: np.ndarray,
        qubit_count: int,
        matrix: np.ndarray,
        targets: tuple[int, ...],
        controls: tuple[int, ...] | None,
        control_state: tuple[int, ...] | None,
        dispatcher: QuantumGateDispatcher,
        gate_type: str | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Apply a unitary gate matrix U to a density matrix \rho according to:

            .. math::
                \rho \rightarrow U \rho U^{\dagger}

        This represents the quantum evolution of a density matrix under a unitary
        operation, where the gate is applied on the left and its Hermitian conjugate
        on the right to preserve the trace and Hermitian properties of the density matrix.

        Args:
            result (np.ndarray): Initial density matrix in reshaped form [2]*(2*qubit_count).
                This buffer may be modified during computation and used for intermediate results.
            temp (np.ndarray): Pre-allocated buffer used for multiply_matrix output operations.
                Must have the same shape and dtype as result.
            qubit_count (int): Number of qubits in the circuit.
            matrix (np.ndarray): Unitary gate matrix U to be applied to the density matrix.
                Will be converted to complex dtype if necessary.
            targets (tuple[int, ...]): Target qubits that the unitary gate acts upon.
            controls (tuple[int, ...] | None): The qubits to control the operation on. Default ().
            control_state (tuple[int, ...] | None): A tuple of same length as `controls` with either
                a 0 or 1 in each index, corresponding to whether to control on the `|0⟩` or `|1⟩` state.
                Default (1,) * len(controls).
            dispatcher (QuantumGateDispatcher): Dispatches multiplying based on qubit count.
            gate_type (str | None): Optional gate type identifier for optimized dispatch.

        Returns:
            tuple[np.ndarray, np.ndarray]: A tuple containing:
                - The output density matrix (U * \rho * U†)
                - A spare buffer that can be reused for subsequent operations

        Note:
            The function uses efficient buffer swapping to minimize memory allocations.
            The shifted targets (targets + qubit_count) are used for the right-side
            multiplication with U† to account for the doubled dimension structure
            of the reshaped density matrix.
        """
        _, needs_swap1 = multiply_matrix(
            state=result,
            matrix=matrix,
            targets=targets,
            controls=controls,
            control_state=control_state,
            out=temp,
            return_swap_info=True,
            dispatcher=dispatcher,
            gate_type=gate_type,
        )
        if needs_swap1:
            result, temp = temp, result

        multiply_matrix(
            state=result,
            # TODO: Fix control slicing for right multiplication
            matrix=DensityMatrixSimulation._get_controlled_matrix(matrix, control_state).conj(),
            targets=tuple(t + qubit_count for t in controls + targets),
            out=temp,
            return_swap_info=True,
            dispatcher=dispatcher,
            # TODO: remove condition once CNot dispatch is fixed
            gate_type=gate_type if len(targets) == 1 else None,
        )
        # Always swap with new gate dispatch
        result, temp = temp, result
        return result, temp

    @staticmethod
    def _get_controlled_matrix(matrix: np.ndarray, ctrl_state: tuple[int, ...]) -> np.ndarray:
        new_matrix = matrix
        for state in ctrl_state:
            identity = np.eye(len(new_matrix))
            new_matrix = (
                block_diag(identity, new_matrix) if state else block_diag(new_matrix, identity)
            )
        return new_matrix

    @staticmethod
    def _apply_kraus(
        result: np.ndarray,
        temp: np.ndarray,
        work_buffer1: np.ndarray,
        work_buffer2: np.ndarray,
        qubit_count: int,
        matrices: list[np.ndarray],
        targets: tuple[int, ...],
        dispatcher: QuantumGateDispatcher,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Apply a list of matrices {E_i} to a density matrix D according to:

            .. math::
                D \rightarrow \\sum_i E_i D E_i^{\dagger}

        This version uses pre-allocated buffers for memory-efficient computation,
        avoiding repeated memory allocations during the Kraus operation loop.

        Args:
            result (np.ndarray): Initial density matrix in reshaped form [2]^(2*qubit_count).
                This buffer is preserved and never modified during computation.
            temp (np.ndarray): Pre-allocated buffer used as accumulator for the final result.
                Must have the same shape and dtype as result.
            work_buffer1 (np.ndarray): Pre-allocated working buffer for intermediate calculations.
                Must have the same shape and dtype as result.
            work_buffer2 (np.ndarray): Pre-allocated working buffer for multiply_matrix output.
                Must have the same shape and dtype as result.
            qubit_count (int): Number of qubits in the circuit.
            matrices (list[np.ndarray]): Kraus operators {E_i} to be applied to the density matrix.
            targets (tuple[int, ...]): Target qubits that the Kraus operators act upon.
            dispatcher (QuantumGateDispatcher): Dispatches multiplying based on quibit count.

        Returns:
            tuple[np.ndarray, np.ndarray]: A tuple containing:
                - The output density matrix (sum_i E_i * D * E_i†)
                - A spare buffer that can be reused for subsequent operations

        Note:
            The input density matrix in `result` is never modified. Each Kraus operator
            E_i is applied to the original density matrix, and the results are accumulated
            in the `temp` buffer to compute the final sum.
        """
        if len(targets) <= 2:
            superop = sum(np.kron(matrix, matrix.conj()) for matrix in matrices)
            targets_new = targets + tuple([target + qubit_count for target in targets])
            _, needs_swap = multiply_matrix(
                result, superop, targets_new, out=temp, return_swap_info=True, dispatcher=dispatcher
            )
            # With gate_type dispatch, swaps won't occur. An optimization would be to do is add matrix matching to avoid general 1q, 2q cases.
            result, temp = temp, result
            return result, temp

        temp.fill(0)
        shifted_targets = tuple(t + qubit_count for t in targets)
        # Targets are always greater than 2 so we never need to check for swaps
        for matrix in matrices:
            current_buffer = result
            output_buffer = work_buffer1
            multiply_matrix(
                state=current_buffer,
                matrix=matrix,
                targets=targets,
                out=output_buffer,
                dispatcher=dispatcher,
            )
            current_buffer, output_buffer = output_buffer, work_buffer2
            multiply_matrix(
                state=current_buffer,
                matrix=matrix.conj(),
                targets=shifted_targets,
                out=output_buffer,
                dispatcher=dispatcher,
            )
            temp += output_buffer
        result, temp = temp, result

        return result, temp



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from functools import lru_cache, singledispatch

import numpy as np

from braket.default_simulator.operation import GateOperation, KrausOperation


def from_braket_instruction(instruction) -> GateOperation | KrausOperation:
    """Instantiates the concrete `GateOperation` or `KrausOperation` object from the
    specified Braket instruction.

    Args:
        instruction: instruction for a circuit specified using the `braket.ir.jacqd` format.
    Returns:
        GateOperation | KrausOperation: instance of the concrete GateOperation or
        KrausOperation class corresponding to the specified instruction.

    Raises:
        NotImplementedError: If no concrete `GateOperation` or `KrausOperation` class has been
            registered for the instruction type.
    """
    return _from_braket_instruction(instruction)


@singledispatch
def _from_braket_instruction(instruction):
    raise NotImplementedError(f"Instruction {instruction} not recognized")


@lru_cache
def pauli_eigenvalues(num_qubits: int) -> np.ndarray:
    """The eigenvalues of Pauli operators and their tensor products.

    Args:
        num_qubits (int): the number of qubits the operator acts on
    Returns:
        np.ndarray: the eigenvalues of a Pauli product operator of the given size
    """
    if num_qubits == 1:
        return np.array([1, -1], dtype=complex)
    return np.concatenate([pauli_eigenvalues(num_qubits - 1), -pauli_eigenvalues(num_qubits - 1)])


def ir_matrix_to_ndarray(matrix: list[list[list[float]]]) -> np.ndarray:
    """Converts a JAQCD matrix into a numpy array.

    Args:
        matrix (list[list[list[float]]]: The IR representation of a matrix

    Returns:
        np.ndarray: The numpy ndarray representation of the matrix
    """
    return np.array([[complex(element[0], element[1]) for element in row] for row in matrix])


def check_matrix_dimensions(matrix: np.ndarray, targets: tuple[int, ...]) -> None:
    """Checks that the matrix is of the correct shape to act on the targets.

    Args:
        matrix (np.ndarray): The matrix to check
        targets (tuple[int, ...]): The target qubits the matrix acts on

    Raises:
        ValueError: If the matrix is not a square matrix or operates on a space
            of different dimension than that generated by the target qubits
    """
    if len(matrix.shape) != 2 or matrix.shape[0] != matrix.shape[1]:
        raise ValueError(f"{matrix} is not a two-dimensional square matrix")

    dimension = 2 ** len(targets)
    if dimension != matrix.shape[0]:
        raise ValueError(
            f"`matrix` operates on space of dimension {matrix.shape[0]} instead of {dimension}"
        )


def check_unitary(matrix: np.ndarray):
    """Checks that the given matrix is unitary.

    Args:
        matrix (np.ndarray): The matrix to check

    Raises:
        ValueError: If the matrix is not unitary
    """
    if not np.allclose(np.eye(len(matrix)), matrix.dot(matrix.T.conj())):
        raise ValueError(f"{matrix} is not unitary")


def check_hermitian(matrix: np.ndarray):
    """Checks that the given matrix is Hermitian.

    Args:
        matrix (np.ndarray): The matrix to check

    Raises:
        ValueError: If the matrix is not Hermitian
    """
    if not np.allclose(matrix, matrix.T.conj()):
        raise ValueError(f"{matrix} is not Hermitian")


def check_cptp(matrices: list[np.ndarray]):
    """Checks that the given matrices define a CPTP map.

    Args:
        matrices (list[np.ndarray]): The matrices to check

    Raises:
        ValueError: If the matrices do not define a CPTP map
    """
    E = sum([np.matmul(matrix.T.conjugate(), matrix) for matrix in matrices])
    if not np.allclose(E, np.eye(*E.shape)):
        raise ValueError(f"{matrices} do not define a CPTP map")
